from __future__ import print_function
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms

class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)

def train(args, model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))

def test(args, model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)
    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset)))

def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
    parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=10, metavar='N',
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')
    args = parser.parse_args()
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=True, download=True,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ])),
        batch_size=args.batch_size, shuffle=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(
        datasets.MNIST('../data', train=False, transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ])),
        batch_size=args.test_batch_size, shuffle=True, **kwargs)


    model = Net().to(device)
    optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)

    for epoch in range(1, args.epochs + 1):
        train(args, model, device, train_loader, optimizer, epoch)
        test(args, model, device, test_loader)


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
from abc import abstractmethod, ABCMeta
import pandas as pd
import sys
import traceback
import random
from py_common_util.common.common_utils import CommonUtils
from py_common_util.zipline.simple_zipline.trade_order import TradeOrder
from py_common_util.zipline.simple_zipline.bar_data import BarData
from py_common_util.common.enhanced_ordered_dict import EnhancedOrderedDict
from py_common_util.zipline.simple_zipline.trade_context import TradeContext
from six import with_metaclass
from logbook import Logger, StreamHandler, FileHandler


class SimpleAbstractStrategy(with_metaclass(ABCMeta)):
    """
    简单的策略回测抽象，参考了zipline的设计思路
    https://github.com/quantopian/zipline/blob/7056a9091423d1465968d21498b18de75ba7c6c6/zipline/finance/metrics/metric.py
    https://github.com/quantopian/zipline/blob/7e642c8e7ef49b2250d47074ce3295a39bb0726d/zipline/finance/ledger.py
    https://github.com/quantopian/zipline/blob/3106d495f0521639b4be6a77f35c194e849e6c1f/zipline/data/benchmarks.py
    """
    @property
    def log(self):
        return self._log

    @log.setter
    def log(self, value):
        self._log = value

    @property
    def trade_context(self):
        return self._trade_context

    @property
    def bar_data(self):
        return self._bar_data

    @property
    def records(self):
        return self._records

    @property
    def trade_order(self):
        return self._trade_order

    @property
    def multiple_order_position_dict(self):
        return self._multiple_order_position_dict

    @property
    def perf(self):
        return self._perf

    @property
    def sink_fn(self):
        return self._sink_fn

    @property
    def sink_out_dict(self):
        return self._sink_out_dict

    def __init__(self):
        StreamHandler(sys.stdout).push_application()
        self._log = Logger(self.__class__.__name__)
        self._perf = pd.DataFrame()
        self._records = EnhancedOrderedDict()
        self._trade_context = TradeContext()
        self._sink_fn = self.sink
        self._sink_out_dict = EnhancedOrderedDict()
        # build bar_data
        self._bar_data = BarData()
        # build trade order
        self._trade_order = TradeOrder()
        # build multiple order position dict
        self._multiple_order_position_dict = EnhancedOrderedDict()

    def add_sink_fn(self, sink_fn):
        """
        增加sink方法，支持分布式远程执行sink
        :param sink_fn: sink方法
        :return:
        """
        self._sink_fn = sink_fn
        return self

    @abstractmethod
    def prepare_data(self):
        # zipline数据默认缓存在：~/.zipline/data/
        return None

    @abstractmethod
    def initialize(self, context):
        # 启动后需要用户干预处理的一次性逻辑
        self.log.info("initialize...")

    @abstractmethod
    def before_trading_start(self, context, data):
        # 每个bar_open之前执行，对日K可选定当天待交易股票，分钟K在每日盘前可以用于初始化数据
        # self.log.info("before_trading_start...")
        pass

    @abstractmethod
    def handle_data(self, context, data):
        # 定时执行，处理当前周期中待处理订单
        # self.log.info("handle_data...")
        pass

    @abstractmethod
    def analyze(self, context, records):
        self.log.info("analyze...")

    def sink(self, value: EnhancedOrderedDict):
        """持久化输出对象"""
        pass

    def run_algorithm(self):
        try:
            self.prepare_data()
            self.initialize(self.trade_context)
            for trade_date in self.trade_context.trade_calendar_list:
                self.trade_context.current_bar_item_day_date = trade_date
                self.trade_context.current_bar_item_kline_date = trade_date
                self.bar_data.set_current_kline_date(trade_date)
                self.before_trading_start(context=self.trade_context, data=self.bar_data)
                self.handle_data(context=self.trade_context, data=self.bar_data)
            self.analyze(context=self.trade_context, records=self.records)
            self.sink_fn(self.sink_out_dict)
            return self.perf
        except Exception as e:
            self.log.error("error occurred while backtesting: %s" % str(e))
            traceback.print_exc()
        return None

    @CommonUtils.print_exec_time
    def process(self):
        return self.run_algorithm()

    def _draw_return_rate_line(self, perf):
        """"画图，可以用于开发调试"""
        import seaborn as sns
        import matplotlib.pyplot as plt
        from matplotlib.dates import DateFormatter
        sns.set_style('darkgrid')
        sns.set_context('notebook')
        ax = plt.axes()
        # # 设置时间显示格式
        years_fmt = DateFormatter('%Y-%m-%d')
        ax.xaxis.set_major_formatter(years_fmt)
        # 让x轴坐标旋转45度
        labels = ax.get_xticklabels()
        plt.setp(labels, rotation=35, horizontalalignment='right')
        # 画出收益率曲线
        sns.lineplot(x='period_close',
                     y='strategy_return',
                     data=perf,
                     label="strategy_return")
        sns.lineplot(x='period_close',
                     y='benchmark_return',
                     data=perf, label="benchmark_return")
        plt.legend(loc='upper left')
        plt.title("return rate of strategy and benchmark")
        plt.xlabel('time')
        plt.ylabel('return rate')
        plt.show()
        # # save plt
        # fig = ax.get_figure()
        # fig.savefig('～/gender_salary.png')

    @staticmethod
    def _random_subset(list, select_num=10, random_seed=10, head_start=0, head_end=2, foot_start=-2, foot_end=None):
        """
        保留头尾各n个元素，其它元素从list中间元素中随机选择指定数量的元素
        :param list: 目标list e.g. list = [i for i in range(189)]
        :param select_num: 指定输出元素数量（包括头尾元素）
        :param random_seed 随机种子，在random_seed相同的情况下返回结果是固定的
        :param head_start 保留前面部分的第head_start到head_end之间的元素，不包含index为head_end的元素
        :param head_end 同head_start
        :param foot_start 保留后面面部分的第foot_start到foot_end之间的元素，不包含index为head_end的元素(foot_end=None除外)
        :param foot_end 为None表示包含最后一个元素
        :return: 选定后的list e.g. _random_subset([3], 10)->[3], _random_subset(list, 10)->[0, 1, 5, 10, 111, 125, 148, 149, 187, 188]
        """
        head_list = list[head_start:head_end]
        foot_list = list[foot_start:foot_end]
        default_element_num = len(head_list) + len(foot_list)
        if list is None or select_num < default_element_num or len(list) <= select_num:
            return list
        indexs = [i for i in range(len(list))]
        mid_list = indexs[head_end:foot_start]
        random.seed(random_seed)
        slice_indexs = random.sample(mid_list, select_num - default_element_num)
        slice_indexs = sorted(slice_indexs, key=int, reverse=False)
        return head_list + [list[i] for i in slice_indexs] + foot_list


if __name__ == '__main__':
    SimpleAbstractStrategy().process()




import argparse

import numpy as np

from py_common_util.nlptoolkit.elmo.bilm.training import train, load_options_latest_checkpoint, load_vocab
from py_common_util.nlptoolkit.elmo.bilm.data import BidirectionalLMDataset


def main(args):
    # load the vocab
    vocab = load_vocab(args.vocab_file, 50)

    # define the options
    batch_size = 128  # batch size for each GPU
    n_gpus = 3

    # number of tokens in training data (this for 1B Word Benchmark)
    n_train_tokens = 768648884

    options = {
     'bidirectional': True,

     'char_cnn': {'activation': 'relu',
      'embedding': {'dim': 16},
      'filters': [[1, 32],
       [2, 32],
       [3, 64],
       [4, 128],
       [5, 256],
       [6, 512],
       [7, 1024]],
      'max_characters_per_token': 50,
      'n_characters': 261,
      'n_highway': 2},
    
     'dropout': 0.1,
    
     'lstm': {
      'cell_clip': 3,
      'dim': 4096,
      'n_layers': 2,
      'proj_clip': 3,
      'projection_dim': 512,
      'use_skip_connections': True},
    
     'all_clip_norm_val': 10.0,
    
     'n_epochs': 10,
     'n_train_tokens': n_train_tokens,
     'batch_size': batch_size,
     'n_tokens_vocab': vocab.size,
     'unroll_steps': 20,
     'n_negative_samples_batch': 8192,
    }

    prefix = args.train_prefix
    data = BidirectionalLMDataset(prefix, vocab, test=False,
                                      shuffle_on_load=True)

    tf_save_dir = args.save_dir
    tf_log_dir = args.save_dir
    train(options, data, n_gpus, tf_save_dir, tf_log_dir)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--save_dir', help='Location of checkpoint files')
    parser.add_argument('--vocab_file', help='Vocabulary file')
    parser.add_argument('--train_prefix', help='Prefix for train files')
    args = parser.parse_args()
    main(args)




import numpy as np

# N is batch size; D_in is input dimension;
# H is hidden dimension; D_out is output dimension.
N, D_in, H, D_out = 64, 1000, 100, 10

# Create random input and output data
x = np.random.randn(N, D_in)
y = np.random.randn(N, D_out)

# Randomly initialize weights
w1 = np.random.randn(D_in, H)
w2 = np.random.randn(H, D_out)

learning_rate = 1e-6
for t in range(500):
    # Forward pass: compute predicted y
    h = x.dot(w1)
    h_relu = np.maximum(h, 0)
    y_pred = h_relu.dot(w2)

    # Compute and print loss
    loss = np.square(y_pred - y).sum()
    print(t, loss)

    # Backprop to compute gradients of w1 and w2 with respect to loss
    grad_y_pred = 2.0 * (y_pred - y)
    grad_w2 = h_relu.T.dot(grad_y_pred)
    grad_h_relu = grad_y_pred.dot(w2.T)
    grad_h = grad_h_relu.copy()
    grad_h[h < 0] = 0
    grad_w1 = x.T.dot(grad_h)

    # Update weights
    w1 -= learning_rate * grad_w1
    w2 -= learning_rate * grad_w2


# coding: utf-8 or # -*- coding: utf-8 -*-
import torch
import numpy as np


class TorchUtils:

    @staticmethod
    def empty_zero_data(shape_channels_height_width=[1, 28, 28], dtype=np.float32):
        """
        返回length=0的数据
        :param shape_channels_height_width:
        e.g. [(np.zeros((0, 1, 28, 28), dtype=np.float32), np.zeros((0, 10), dtype=np.int64))]
        :return:
        """
        return np.zeros((0,
                        shape_channels_height_width[0],
                        shape_channels_height_width[1],
                        shape_channels_height_width[2]), dtype=dtype)

    @staticmethod
    def watch_tensor(tensor):
        """
        在cpu环境下查看PyTorch的tensor
        """
        return tensor.data.cpu().numpy()

    @staticmethod
    def batchify(tensor_data, batch_size, device):
        """
        将数据按batch_size分批
        :param tensor_data:
        :param batch_size:
        :param device:
        :return:
        """
        num_batch = tensor_data.size(0) // batch_size
        # Trim off any extra elements that wouldn't cleanly fit (remainders).
        tensor_data = tensor_data.narrow(0, 0, num_batch * batch_size)
        # Evenly divide the data across the batch_size batches.
        tensor_data = tensor_data.view(batch_size, -1).t().contiguous()
        return tensor_data.to(device)

    @staticmethod
    def total_num_batch(num_total, batch_size, allow_smaller_final_batch=False):
        num_batch = num_total // batch_size
        if allow_smaller_final_batch:
            if num_total > (num_batch * batch_size):
                return num_batch + 1
        else:
            # 舍去最后不能够成1批的剩余训练数据部分
            return num_batch

    @staticmethod
    def batchify_np_data_label(tuple_np_data_label, batch_size, allow_smaller_final_batch=False):
        """
        :param tuple_np_data_label: [([1,2,3],[0,0,1])]
        :param batch_size: 128
        :return: [(1,0),(2,0),(3,1)]
        使用：
        ......
        batched_data_label = TorchUtils.batchify_np_data_label(np_test_data, parser_args.eval_batch_size)
        with torch.no_grad():
            for batch_idx, (data, label) in enumerate(batched_data_label):
                data, label = TorchUtils.np_to_tensor(data).to(device), TorchUtils.np_to_tensor(label).to(device)
        ......
        """
        data = tuple_np_data_label[0][0]
        label = tuple_np_data_label[0][1]
        num_batch = TorchUtils.total_num_batch(data.shape[0], batch_size, allow_smaller_final_batch)
        data = data[:num_batch*batch_size]
        label = label[:num_batch*batch_size]
        batched_data = np.split(data, num_batch)
        batched_label = np.split(label, num_batch)
        return list(zip(batched_data, batched_label))

    @staticmethod
    def np_to_tensor(np_data):
        return np_data if TorchUtils.is_tensor(np_data) else torch.from_numpy(np_data)

    @staticmethod
    def is_tensor(tensor_data):
        return torch.is_tensor(tensor_data)

    @staticmethod
    def cuda_is_available():
        return torch.cuda.is_available()

    @staticmethod
    def device(device="gpu", device_index=0):
        return torch.device("cuda" if device=='gpu' else "cpu", device_index)


if __name__ == '__main__':
    data = torch.randn(2, 5)
    print(data)



# -*- coding: utf-8 -*-
import uuid
from py_common_util.common.enhanced_ordered_dict import EnhancedOrderedDict


class MultipleOrderPosition(object):
    """
    多个标的每日轮动回测类型的订单，参考：https://github.com/quantopian/zipline/blob/master/zipline/finance/order.py
    每个调仓周期只对应1个MultipleOrderPosition对象，而每个MultipleOrderPosition对象对应有调仓周期中所有日期的持仓字典position_dict
    init_cash: 回测时的初始总金额。后面会浮动变成total_value
    total_profit 多只股票的总收益 (会根据最新价浮动变化，每笔交易中多只股票里累计每只股票的股数*price，不计算手续费和min_move)
    total_balance_cash 【start_date调仓后 至 end_date日准备进行下个调仓周期前的持仓】剩余的总现金 (不会根据最新价浮动变化，本金减去了每笔历史交易当时的total_profit、手续费、min_move，之后还剩余的现金)
    total_value  每日策略内多只股票累加的总价值=total_profit+total_balance_cash
    turnover_ratio 调仓换手率。调仓日换手率计算公式:（每次调仓日卖出的股票市值+每次调仓日买入的股票市值 * 0.5 / (调仓后）账户总资产
    """
    @property
    def position_dict(self):
        return self._position_dict

    def __init__(self, start_date, init_cash, id=None):
        self.id = self._make_id() if id is None else id
        self.start_date = start_date  # 格式如："2019-01-01"
        self.end_date = ""
        self.total_profit = 0.0
        self.total_balance_cash = init_cash
        self.total_value = init_cash
        self._position_dict = EnhancedOrderedDict()  # 每日日期-> trade_order.position_dict.copy()
        self.turnover_adjust_total_value = 0.0  # 每次调仓时卖出和调整过的股票市值
        self.turnover_ratio = 0.0  # 调仓换手率

    def set_end_date(self, end_date):
        """更新该调仓周期的结束日期"""
        self.end_date = end_date

    def set_turnover_adjust_total_value(self, turnover_adjust_total_value):
        """在新的调仓日期调仓完毕后更新该调仓周期的调仓换手率之计算分子：每次调仓时卖出和调整过的股票市值"""
        self.turnover_adjust_total_value = turnover_adjust_total_value

    def get_last_day_position_dict(self):
        """
        使用：[security_code for security_code in get_last_day_position_dict()[1].keys()]
        :return 返回tuple: (trade_date, security_position_dict)
        """
        if len(self.position_dict) > 0:
            return self.position_dict.to_list()[-1]
        else:
            return None

    def update(self, trade_date, position_dict_copy, total_balance_cash):
        # 更新每天的持仓明细
        self.position_dict[trade_date] = position_dict_copy
        self.total_balance_cash = total_balance_cash
        total_profit = 0
        should_del_security_list = []
        for security_code in self.position_dict[trade_date]:
            position = self.position_dict[trade_date].get(security_code)
            if position.amount > 0:
                total_profit += position.amount * position.last_price
                self.total_balance_cash += position.balance_cash
            else:
                should_del_security_list.append(security_code)
        # 删除amount=0的已经卖空了的持仓
        for security_code in should_del_security_list:
            del self.position_dict[trade_date][security_code]
        self.total_profit = total_profit
        self.total_value = self.total_profit + self.total_balance_cash
        # 计算每个调仓日的换手率
        if self.start_date == trade_date and self.total_value > 0:
            self.turnover_ratio = self.turnover_adjust_total_value * 0.5 / self.total_value

    @staticmethod
    def _make_id():
        return uuid.uuid4().hex

    def to_dict(self):
        """
        Creates a dictionary representing the state of this position.
        Returns a dict object of the form:
        """
        position_dict_str = ""
        for trade_date in self.position_dict:
            position_dict_str += trade_date + ":"
            for security_code in self.position_dict.get(trade_date):
                position_dict_str += str(self.position_dict.get(trade_date).get(security_code).to_dict()) + ","
        return {
            'start_date': self.start_date,
            'end_date': self.end_date,
            'total_balance_cash': self.total_balance_cash,
            "total_profit": self.total_profit,
            "total_value": self.total_value,
            "turnover_adjust_total_value": self.turnover_adjust_total_value,
            "turnover_ratio": self.turnover_ratio,
            "position_dict": position_dict_str
        }



# -*- coding: utf-8 -*-
from collections import deque


class RingBuffer(deque):
    """
    python的ring buffer高性能封装
    注意：deque性能远高于封装的RingBuffer类，因此应尽量使用原始的deque对象
    为性能考虑，应注释一切代理的方法(i.e. append, get)，程序中应直接使用deque原始支持的方法
    参考：https://en.wikipedia.org/wiki/Circular_buffer
    # def append(self, x):
        # super().append(x)
    """
    def __init__(self, maxlen):
        super().__init__(maxlen=maxlen)


if __name__ == '__main__':
    import time
    tm1 = time.time()
    rb = RingBuffer(1000)
    for i in range(int(1e6)):
        rb.append(i)
    print(list(rb)[:10])
    rb.clear()
    tm2 = time.time()
    print("使用封装的ringbuffer执行时间；{:.2f}seconds".format(tm2 - tm1))

    import time
    import collections
    tm1 = time.time()
    d = collections.deque(maxlen=1000)
    for i in range(int(1e6)):
        d.append(i)
    print(list(d)[:10])
    d.clear()
    tm2 = time.time()
    print("使用原始的deque执行时间；{:.2f}seconds".format(tm2 - tm1))



#  Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""An Example of a custom Estimator for the Iris dataset."""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function

import argparse
import tensorflow as tf

from py_common_util.tensorflow.sample import iris_data

parser = argparse.ArgumentParser()
parser.add_argument('--batch_size', default=100, type=int, help='batch size')
parser.add_argument('--train_steps', default=1000, type=int,
                    help='number of training steps')

"""
预创建的 Estimator https://www.tensorflow.org/get_started/premade_estimators
"""


def my_model(features, labels, mode, params):
    """DNN with three hidden layers, and dropout of 0.1 probability."""
    # Create three fully connected layers each layer having a dropout
    # probability of 0.1.
    net = tf.feature_column.input_layer(features, params['feature_columns'])
    for units in params['hidden_units']:
        net = tf.layers.dense(net, units=units, activation=tf.nn.relu)

    # Compute logits (1 per class).
    logits = tf.layers.dense(net, params['n_classes'], activation=None)

    # Compute predictions.
    predicted_classes = tf.argmax(logits, 1)
    if mode == tf.estimator.ModeKeys.PREDICT:
        predictions = {
            'class_ids': predicted_classes[:, tf.newaxis],
            'probabilities': tf.nn.softmax(logits),
            'logits': logits,
        }
        return tf.estimator.EstimatorSpec(mode, predictions=predictions)

    # Compute loss.
    loss = tf.losses.sparse_softmax_cross_entropy(labels=labels, logits=logits)

    # Compute evaluation metrics.
    accuracy = tf.metrics.accuracy(labels=labels,
                                   predictions=predicted_classes,
                                   name='acc_op')
    metrics = {'accuracy': accuracy}
    tf.summary.scalar('accuracy', accuracy[1])

    if mode == tf.estimator.ModeKeys.EVAL:
        return tf.estimator.EstimatorSpec(
            mode, loss=loss, eval_metric_ops=metrics)

    # Create training op.
    assert mode == tf.estimator.ModeKeys.TRAIN

    optimizer = tf.train.AdagradOptimizer(learning_rate=0.1)
    train_op = optimizer.minimize(loss, global_step=tf.train.get_global_step())
    return tf.estimator.EstimatorSpec(mode, loss=loss, train_op=train_op)


def main(argv):
    args = parser.parse_args(argv[1:])

    # Fetch the data
    (train_x, train_y), (test_x, test_y) = iris_data.load_data()

    # Feature columns describe how to use the input.
    my_feature_columns = []
    for key in train_x.keys():
        my_feature_columns.append(tf.feature_column.numeric_column(key=key))

    # Build 2 hidden layer DNN with 10, 10 units respectively.
    classifier = tf.estimator.Estimator(
        model_fn=my_model,
        params={
            'feature_columns': my_feature_columns,
            # Two hidden layers of 10 nodes each.
            'hidden_units': [10, 10],
            # The model must choose between 3 classes.
            'n_classes': 3,
        })

    # Train the Model.
    classifier.train(
        input_fn=lambda: iris_data.train_input_fn(train_x, train_y, args.batch_size),
        steps=args.train_steps)

    # Evaluate the model.
    eval_result = classifier.evaluate(
        input_fn=lambda: iris_data.eval_input_fn(test_x, test_y, args.batch_size))

    print('\nTest set accuracy: {accuracy:0.3f}\n'.format(**eval_result))

    # Generate predictions from the model
    expected = ['Setosa', 'Versicolor', 'Virginica']
    predict_x = {
        'SepalLength': [5.1, 5.9, 6.9],
        'SepalWidth': [3.3, 3.0, 3.1],
        'PetalLength': [1.7, 4.2, 5.4],
        'PetalWidth': [0.5, 1.5, 2.1],
    }

    predictions = classifier.predict(
        input_fn=lambda: iris_data.eval_input_fn(predict_x,
                                                 labels=None,
                                                 batch_size=args.batch_size))

    for pred_dict, expec in zip(predictions, expected):
        template = ('\nPrediction is "{}" ({:.1f}%), expected "{}"')

        class_id = pred_dict['class_ids'][0]
        probability = pred_dict['probabilities'][class_id]

        print(template.format(iris_data.SPECIES[class_id],
                              100 * probability, expec))


if __name__ == '__main__':
    tf.logging.set_verbosity(tf.logging.INFO)
    tf.app.run(main)


# -*- coding: utf-8 -*-
from py_common_util.zipline.chinese_stock_calendar import ChineseStockCalendar
from py_common_util.zipline.default_strategy import DefaultStrategy
from py_common_util.common.common_utils import CommonUtils
from empyrical import cum_returns, annual_return, sharpe_ratio, max_drawdown, alpha, beta
import pandas as pd
from collections import OrderedDict
from trading_calendars import get_calendar


class NocodeBacktestZiplineStrategy(DefaultStrategy):
    """
    无编程回测zipline策略类
    """
    def __init__(self):
        super().__init__()

    def prepare_data(self):
        """
        AAPL.csv数据涞源：https://finance.yahoo.com/quote/AAPL/history/
        :return:
        """
        super().prepare_data()
        data_dict = OrderedDict()
        data = {
                'trade_date': pd.Series([self.pandas.Timestamp('2018-07-30 00:00:00', tz='utc'),
                                         self.pandas.Timestamp('2018-07-31 00:00:00', tz='utc'),
                                         self.pandas.Timestamp('2018-08-01 00:00:00', tz='utc')]),
                'open': pd.Series([1.0, 2.0, 1.0]),
                'high': pd.Series([1.0, 2.0, 1.0]),
                'low': pd.Series([1.0, 2.0, 1.0]),
                'close': pd.Series([201, 200, 203]),
                'volume': pd.Series([1.0, 2.0, 1.0])
                }
        data_dict["SPY"] = pd.DataFrame(data)
        data_dict["SPY"].set_index("trade_date", inplace=True)
        data_dict["SPY"]["benchmark_return"] = data_dict["SPY"]["close"].pct_change()
        data_dict["SPY"]["benchmark_return"].fillna(0, inplace=True)
        data_dict['SPY'] = data_dict['SPY'][['open', 'high', 'low', 'close', 'volume',"benchmark_return"]]
        print(data_dict["SPY"].head())
        panel = self.pandas.Panel(data_dict)
        panel.major_axis = panel.major_axis.tz_convert(tz='utc')
        panel.minor_axis = ['open', 'high', 'low', 'close', 'volume', "benchmark_return"]
        return panel

    def initialize(self, context):
        self.log.info("sid(int)=" + str(self.zipline.api.sid(0)))  # 这里的sid(0)等价symbol('SPY')
        context.sym = self.zipline.api.symbol('SPY')
        self.zipline.api.set_benchmark(context.sym)
        context.i = 0
        context.set_commission(self.zipline.finance.commission.PerShare(cost=.0075, min_trade_cost=1.0))
        context.set_slippage(self.zipline.finance.slippage.VolumeShareSlippage())

    def before_trading_start(self, context, data):
        self.log.info("before_trading_start...")
        pass

    def handle_data(self, context, data):
        self.log.info("handle_data...")
        context.i += 1
        # if context.i < 300:
        #     return
        # self.zipline.api.order(asset=context.sym, amount=100, limit_price=10.23)
        asset = self.zipline.api.symbol("SPY")
        self.zipline.api.order(asset, amount=100, limit_price=201.7)
        # self.zipline.api.order_target(context.sym, 123)
        self.zipline.api.order_target_percent(context.sym, 0.5)

    def analyze(self, context, records):
        self.log.info("analyze...")
        pass

    @CommonUtils.print_exec_time
    def run_algorithm(self):
        us_calendar = get_calendar("XNYS")
        cn_calendar = ChineseStockCalendar(data_frequency="daily")
        hk_calendar = get_calendar("XHKG")
        start_time = self.pandas.Timestamp('2018-07-30 09:31:00', tz='utc')
        end_time = self.pandas.Timestamp('2018-08-01 16:00:00', tz='utc')
        perf = self.zipline.run_algorithm(start=start_time,
                                          end=end_time,
                                          initialize=self.initialize,
                                          capital_base=100000,
                                          handle_data=self.handle_data,
                                          before_trading_start=self.before_trading_start,
                                          data_frequency="daily",
                                          data=self.prepare_data(),
                                          trading_calendar=us_calendar,
                                          analyze=self.analyze)
        return perf


def draw_return_rate_line(result):
    import matplotlib.pyplot as plt
    from matplotlib.dates import DateFormatter
    import seaborn as sns
    sns.set_style('darkgrid')
    sns.set_context('notebook')
    ax = plt.axes()
    # 设置时间显示格式
    years_fmt = DateFormatter('%Y-%m-%d')
    ax.xaxis.set_major_formatter(years_fmt)
    # 让x轴坐标旋转45度
    labels = ax.get_xticklabels()
    plt.setp(labels, rotation=35, horizontalalignment='right')
    # 画出收益率曲线
    sns.lineplot(x='period_close',
    y='algorithm_period_return',
    data=result,
    label="AAPL")
    sns.lineplot(x='period_close',
        y='benchmark_period_return',
        data=result, label="SPX")
    plt.legend(loc='upper left')
    plt.title("return rate of AAPL and SPX")
    plt.xlabel('time')
    plt.ylabel('return rate')
    plt.show()


if __name__ == '__main__':
    # BenchmarkReturnsAndVolatility().start_of_simulation()
    perf = NocodeBacktestZiplineStrategy().run_algorithm()
    print(perf)
    print("========")
    # 画出收益曲线图
    draw_return_rate_line(perf)
    return_list = perf['returns']
    # 计算年化收益率
    ann_return = annual_return(return_list)
    # 计算累计收益率
    cum_return_list = cum_returns(return_list)
    # 计算sharp ratio
    sharp = sharpe_ratio(return_list)
    # 最大回撤
    max_drawdown_ratio = max_drawdown(return_list)
    print("年化收益率 = {:.2%}, 累计收益率 = {:.2%}, 最大回撤 = {:.2%}, 夏普比率 = {:.2f} ".format
          (ann_return, cum_return_list[-1], max_drawdown_ratio, sharp))

    returns = pd.Series(
        index=pd.date_range('2017-03-10', '2017-03-19'),
        data=(-0.012143, 0.045350, 0.030957, 0.004902, 0.002341, -0.02103, 0.00148, 0.004820, -0.00023, 0.01201)
    )
    benchmark_returns = pd.Series(
        index=pd.date_range('2017-03-10', '2017-03-19'),
        data=(-0.031940, 0.025350, -0.020957, -0.000902, 0.007341, -0.01103, 0.00248, 0.008820, -0.00123, 0.01091)
    )
    alpha_return = alpha(returns=returns, factor_returns=benchmark_returns, risk_free=0.01)
    beta_return = beta(returns=returns, factor_returns=benchmark_returns, risk_free=0.01)
    print("alpha_return", alpha_return)
    print("\nbeta_return", beta_return)
    ###############
    import numpy as np
    from empyrical import alpha_beta
    returns = np.array([.01, .02, .03, -.4, -.06, -.02])
    benchmark_returns = np.array([.02, .02, .03, -.35, -.05, -.01])
    # calculate the max drawdown
    max_drawdown(returns)
    # calculate alpha and beta
    alpha, beta = alpha_beta(returns, benchmark_returns)
    print("*********")
    print(alpha, beta)







# -*- coding: utf-8 -*-


class RegularRotationTradeMode(object):
    """定期轮动回测交易模式"""
    def __init__(self, rebalance_cycle:int=1, max_holding_count:int=10, buy_way:int=0):
        self.rebalance_cycle = int(rebalance_cycle)   # 调仓日期, 如: 5, 即第1天，第6天，第11天，。。。
        self.max_holding_count = int(max_holding_count)  # 最大持仓股票个数
        self.buy_way = int(buy_way)  # 买入方式 0:等权重，下标从0开始


# -*- coding:utf-8 -*-
import csv
import copy
import math


class CommonUtils:

    @staticmethod
    def write_csv(file_name, mode='w', write_fn=None):
        """
        write record to csv file
        :param file:
        :param mode: 'r','w'
        :param csvwrite_fn:
        :return:
        """
        with open(file_name, mode) as stream:
            csvwriter = csv.writer(stream)
            write_fn(csvwriter)

    @staticmethod
    def write_h5(file_name, mode='w', db_name='', shape=(1,), dtype='f', data=None):
        import h5py
        h5_file = h5py.File(file_name, mode)
        h5_dataset = h5_file.create_dataset(db_name, shape, data)
        return h5_file

    @staticmethod
    def cycle(iterable=[None]):
        """
        e.g.
        from py_common_util.common.common_utils import CommonUtils
        ...
        iter = CommonUtils.cycle([0,2,3,1])
        next(iter)->0,next(iter)->2,...,next(iter)->0,next(iter)->2,...
        :param iterable: e.g. [0,2,3,1]
        :return:
        """
        from itertools import cycle
        return cycle(iterable)

    @staticmethod
    def datetime_to_int(datetime):
        import time
        return int(time.mktime(datetime.timetuple()))

    @staticmethod
    def int_to_datetime(datetime_int):
        from datetime import datetime
        return datetime.fromtimestamp(datetime_int)

    @staticmethod
    def str_to_datetime(datetime_str, format="%Y%m%d%H%M%S"):
        from datetime import datetime
        return datetime.strptime(datetime_str, format)

    @staticmethod
    def datetime_to_str(date_time, format="%Y%m%d%H%M%S"):
        from datetime import datetime
        return datetime.strftime(date_time, format)  # 将日期转成字面整型字符串, e.g. date: 2009-12-08 16:34:00 -> '20091208163400'

    @staticmethod
    def datetime_now():
        from datetime import datetime
        return datetime.now()

    @staticmethod
    def camelcase_to_snakecase(class_name):
        """
        Convert camel-case string to snake-case.
        e.g. SuperDatasetBuilder.camelcase_to_snakecase(RecommendDatasetBuilder().__class__.__name__)
        -> "recommend_dataset_builder"
        or SuperDatasetBuilder.camelcase_to_snakecase("RecommendDatasetBuilder")
        -> "recommend_dataset_builder"
        """
        # @see tensorflow_datasets.core.naming#camelcase_to_snakecase
        import re
        _first_cap_re = re.compile("(.)([A-Z][a-z0-9]+)")
        _all_cap_re = re.compile("([a-z0-9])([A-Z])")
        s1 = _first_cap_re.sub(r"\1_\2", class_name)
        return _all_cap_re.sub(r"\1_\2", s1).lower()

    @staticmethod
    def print_exec_time(func):
        """
        装饰器：print execution time
        usage:
        @print_exec_time
        def add(x, y=10):
            return x + y
        if __name__ == '__main__':
        add(1, 2) -> "func: add(), time taken: 0.000002 seconds"
        :param func:
        :return:
        """
        def decorator_func(*args, **kwargs):
            from time import time
            before_time = time()
            rv = func(*args, **kwargs)
            print('print_exec_time->func: '+func.__name__+'()' + ', time taken: {:.3f} seconds'.format(time() - before_time))
            return rv
        return decorator_func

    @staticmethod
    def deepcopy(x, memo=None, _nil=[]):
        """深拷贝，该方法执行比较耗时间"""
        return copy.deepcopy(x, memo, _nil)

    @staticmethod
    def format_number(x, times=1, format_args="0.2f"):
        """
        格式化数字输出, e.g. 12.3456-> "12.34"
        :param x: 数字 e.g. 12.3456
        :param times: 乘以的倍数，方便%输出
        :param format_args: e.g. "0.2f" 保留小数点后2位输出
        :return: "12.34" 如果要to float类型就用: float(format_number(12.3456, 100))
        """
        if x is None or math.isnan(x):
            return "0.00"
        try:
            return format(float(x)*times, format_args)
        except:
            return str(x)





# -*- coding:utf-8 -*-
import time
from datetime import datetime
import datetime as dt


class DateUtils:

    @staticmethod
    def to_int(datetime):
        # 转换成时间秒, 用time.time()可以返回有小数点如0.1的秒整数
        return int(time.mktime(datetime.timetuple()))

    @staticmethod
    def int_to_datetime(datetime_int):
        return datetime.fromtimestamp(datetime_int)

    @staticmethod
    def str_to_datetime(datetime_str, format="%Y%m%d%H%M%S"):
        return datetime.strptime(datetime_str, format)

    @staticmethod
    def to_str(date_time, format="%Y%m%d%H%M%S"):
        """
        e.g. DateUtils.to_str(DateUtils.now(), format="%Y-%m-%d %H:%M:%S.%f")[:-3] = "2019-01-22 00:49:25.216"
        :param date_time: date_time类型， 如果是字符串类型则直接返回
        :param format: "%Y%m%d%H%M%S.%f" 格式为 "年月日时分秒.6位毫秒" 经[:-3]可以变为3位毫秒
        :return:
        """
        return date_time.strftime(format)  # 转成字面整型字符串, e.g. date: 2009-12-08 16:34:00 -> '20091208163400'

    @staticmethod
    def now():
        return datetime.now()

    @staticmethod
    def now_to_int():
        return DateUtils.to_int(DateUtils.now())

    @staticmethod
    def now_to_str():
        """当前时间转成string格式输出，如：2019-07-24 09:19:09.668136"""
        return DateUtils.to_str(DateUtils.now(), format="%Y-%m-%d %H:%M:%S.%f")

    @staticmethod
    def calc_duration_seconds(start_time_int, end_time_int):
        """
        :param start_time_int:
        :param end_time_int:
        :return:
        usage:
            start_time_int = DateUtils.to_int(DateUtils.now())
            DateUtils.sleep(5)
            duration = DateUtils.calc_duration_seconds(start_time_int, DateUtils.to_int(DateUtils.now()))
            print(duration)
        """
        return end_time_int - start_time_int

    @staticmethod
    def calc_delta_days(start_date=None, end_date=None, include_start_end_date=False):
        """
        计算从start_date到end_date的所有日期数组（包含start_date, end_date）
        """
        timedelta = end_date - start_date
        if include_start_end_date:
            return [start_date + dt.timedelta(days=int(i)) for i in range(0, timedelta.days+1)]
        else:
            return [start_date + dt.timedelta(days=int(i)) for i in range(1, timedelta.days)]

    @staticmethod
    def sleep(seconds):
        time.sleep(seconds)


if __name__ == '__main__':
    print(DateUtils.to_str(DateUtils.now(), format="%Y-%m-%d %H:%M:%S.%f")[:-3])
    start_date = DateUtils.str_to_datetime("20180302", "%Y%m%d")
    print(DateUtils.calc_delta_days(start_date, DateUtils.now()))
    duration = DateUtils.calc_duration_seconds(DateUtils.now_to_int(), DateUtils.to_int(DateUtils.str_to_datetime('20190109234900')))
    print(duration)



# coding: utf-8 or # -*- coding: utf-8 -*-
"""
NAACL2018 一种新的embedding方法--原理与实验 Deep contextualized word representations (ELMo)
https://cstsunfu.github.io/2018/06/ELMo/
"""
from allennlp.commands.elmo import ElmoEmbedder
elmo = ElmoEmbedder(options_file='/Users/tony/myfiles/spark/share/python-projects/deep_trading/dataset/elmo_embedder/elmo_options.json',
                    weight_file='/Users/tony/myfiles/spark/share/python-projects/deep_trading/dataset/elmo_embedder/elmo_2x4096_512_2048cnn_2xhighway_weights.hdf5',
                    cuda_device=-1)
context_tokens = [['I', 'love', 'you', '.'], ['Sorry', ',', 'I', 'don', "'t", 'love', 'you', '.']]
elmo_embedding, elmo_mask = elmo.batch_to_embeddings(context_tokens)
print(elmo_embedding)
print(elmo_mask)


# -*- coding: utf-8 -*-
import pandas as pd
import json
import traceback
from cassandra.cluster import Cluster
from cassandra.policies import DCAwareRoundRobinPolicy
from rediscluster import StrictRedisCluster
from py_common_util.common.common_utils import CommonUtils


class BarData(object):
    """bar数据，参考zipline._protocol.BarData"""
    @property
    def current_kline_date(self):
        return self._current_kline_date

    @property
    def cassandra_session(self):
        return self._cassandra_session

    @property
    def simple_cassandra_session(self):
        return self._simple_cassandra_session

    def __init__(self):
        self._current_kline_date = None   # e.g. "2019-07-24"
        self._simple_cassandra_session = None
        self._cassandra_session = None

    def do_init(self, stock_type, cassandra_host_port, cassandra_key_space, redis_conn_nodes):
        self._stock_type = stock_type  # e.g. 'HK', 'US'
        # init cassandra session
        def pandas_factory(colnames, rows):
            return pd.DataFrame(rows, columns=colnames)
        cassandra_host, cassandra_port = cassandra_host_port.split(":")
        cluster = Cluster([cassandra_host],
                          load_balancing_policy=DCAwareRoundRobinPolicy(local_dc="datacenter1"),
                          port=int(cassandra_port))
        self._simple_cassandra_session = cluster.connect(cassandra_key_space)  # 简单的连接模式
        self._cassandra_session = cluster.connect(cassandra_key_space)
        self._cassandra_session.default_fetch_size = 10000000  # needed for large queries, otherwise driver will do pagination. Default is 50000.
        self._cassandra_session.row_factory = pandas_factory
        self._redis_conn_nodes = redis_conn_nodes

    def set_current_kline_date(self, current_kline_date):
        self._current_kline_date = current_kline_date

    def can_trade(self, security_code):
        """
        当前bar能否交易
        :param security_code: 交易标的 e.g. "00700.HK"
        :param kline_date: bar日期 e.g. "2019-07-24"
        :return:
        """
        return True

    def current(self, security_code_list):
        """
        获取当前bar的itemView信息，重试10次
        :param security_code_list, e.g. ["00700.HK","01800.HK]
        :param kline_date: bar日期 e.g. "2019-07-24"
        :return: bar_item e.g. {"security_code": "00700.HK", "close": 1.0, "kline_date": "2019-07-24"}
        """
        return self._daily_bar_to_pandas(cassandra_session=self.simple_cassandra_session,
                                         security_code_list=security_code_list,
                                         kline_date=self.current_kline_date)

    def calc_lot_size(self, security_code):
        """一手股票的股数, 美股为1，A股为100，港股中每手的股数不同"""
        if ".HK" in security_code:
            redis_key = "spark:data:hq_security_definition:vo:" + security_code
            redis_client = StrictRedisCluster(startup_nodes=self._redis_conn_nodes, decode_responses=True)
            lot_size = 100
            try:
                result = json.loads(redis_client.get(redis_key))
                lot_size = result["lotSize"]
            except Exception as e:
                print("error occurred at calc_lot_size for get hk lot size from redis: %s" % str(e))
            return lot_size
        elif ".SZ" in security_code or ".SH" in security_code:
            return 100
        else:
            return 1

    def can_buy_lot(self, cash, lot_size, price:float, min_move, commision):
        """
        计算当前金额可以买多少股票
        假设条件：1）当前腾讯股价为100CNY/股，2）有最小买入单位“手”=100股
        此时已经确定的逻辑：腾讯一手价值10000CNY，最多能持有19手（19手价值190000CNY<20万CNY，20手价值200000CNY=20万CNY，但是因为有手续费，所以不能采纳），手续费标准为千一，实际为190CNY。
        则完成该仓位时，分配给腾讯的仓位应该为19手腾讯（股票）+（10000-190）CNY现金。
        :param cash: e.g. 20万
        :param lot_size: e.g. 100
        :param price: e.g. 100
        :param min_move: e.g. 0
        :param commision: e.g. 0.001
        :return:
        """
        return self._calc_trade_lot(cash / (price + min_move + commision), lot_size)

    def _calc_trade_lot(self, trade_lot:int, lot_size:int=1):
        """
        把计划交易的股数转成可以被手数整除的实际股数
        :param trade_lot 计划交易的股数， 可以为float类型的值
        :param lot_size 1手多少股
        """
        # if trade_lot <= lot_size:  # 如果数量小于1手，就下1手的数量
        #     return lot_size
        if trade_lot <= lot_size:  # 如果数量小于1手，就下0手的数量
            return 0
        if trade_lot % lot_size == 0:  # 刚好整除就下trade_lot数量，否则向下取整
            return trade_lot
        else:
            return int(trade_lot / lot_size) * lot_size

    # @CommonUtils.print_exec_time
    def _daily_bar_to_pandas(self, cassandra_session, security_code_list, kline_date) -> pd.DataFrame:
        """
        参考：Get a Pandas DataFrame from a Cassandra query  https://gist.github.com/gioper86/b08b72d77c4e0aefa0137fc3655488dd
        https://stackoverflow.com/questions/41247345/python-read-cassandra-data-into-pandas
        :return:
        """
        table_name = ''
        if self._stock_type == 'HK':
            table_name = 'nocode_quant_hk_screen_data'
        elif self._stock_type == 'US':
            table_name = 'nocode_quant_us_screen_data'
        security_code_list_str = "'" + "','".join(security_code_list) + "'"
        filter_sql = """
        select security_code, hfq_close as close 
        from {0} 
        where trade_date = {1}
        and security_code in ({2})
        """.format(table_name, "'" + kline_date + "'", security_code_list_str)
        # 手工调用cassandra to pandas
        try:
            rows = cassandra_session.execute(filter_sql)
            security_list = []
            close_list = []
            for row in rows:
                security_list.append(row[0])
                close_list.append(row[1])
            data = {
                'security_code': pd.Series(security_list),
                'close': pd.Series(close_list),
            }
            return pd.DataFrame(data)
        except Exception as e:
            print("bar_data#_daily_bar_to_pandas.filter_sql=%s" % filter_sql)
            print("bar_data#_daily_bar_to_pandas error occurred: %s" % str(e))
            traceback.print_exc()
        return None




# -*- coding: utf-8 -*-
from decimal import Decimal
from sys import float_info
import numpy as np
import math
from py_common_util.common.enhanced_ordered_dict import EnhancedOrderedDict
from py_common_util.zipline.simple_zipline.position import Position


class TradeOrder(object):
    """
    订单处理, 参考：zipline.TradingAlgorithm
    https://github.com/quantopian/zipline/blob/master/zipline/finance/ledger.py
    https://github.com/quantopian/zipline/blob/master/zipline/finance/order.py
    """
    @property
    def benchmark_security_code(self):
        return self._benchmark_security_code

    @property
    def bar_data(self):
        return self._bar_data

    @property
    def min_move(self):
        return self._min_move

    @property
    def commission(self):
        return self._commission

    @property
    def position_dict(self):
        return self._position_dict

    @property
    def multiple_order_position_dict(self):
        return self._multiple_trade_order_dict

    def __init__(self):
        self._benchmark_security_code = None
        self._bar_data = None
        self._commission = None
        self._min_move = 0.0  # TODO default value is 0.01， 目前让其为0
        self._position_dict = EnhancedOrderedDict()  # 每只股票代码->1个Position对象

    def do_init(self, bar_data, commission):
        self._bar_data = bar_data
        self._commission = commission

    def reinit_position_dict(self, target_security_init_cash_list):
        """
        重新初始化所有股票的持仓
        @:param target_security_init_cash_list e.g. {"00700.HK": {"init_cash_per_stock":333333, "balance_cash":0}}
        """
        for target_security_code in target_security_init_cash_list.keys():
            new_init_cash = target_security_init_cash_list[target_security_code]["init_cash_per_stock"]
            balance_cash = target_security_init_cash_list[target_security_code]["balance_cash"]
            target_position = self.position_dict.get(target_security_code)
            if target_position:
                self.position_dict.get(target_security_code).reinit(init_cash_per_stock=new_init_cash,
                                                                    balance_cash=balance_cash,
                                                                    amount=target_position.amount,
                                                                    start_price=0)
            else:
                self.position_dict[target_security_code] = Position(target_security_code, init_cash_per_stock=new_init_cash)
        # 删除非target_security_init_cash_list里的所有position
        keys = [key for key in self.position_dict.keys()]
        for security_code in keys:
            if security_code not in target_security_init_cash_list:
                del self.position_dict[security_code]

    def order_target(self,
                     security_code,
                     target,
                     limit_price=None,
                     stop_price=None,
                     style=None):
        """
        参考：zipline.api.order_target()
        :return 返回提交委托交易的股票数量和价格 e.g. (1, 33.34)
        """
        if not self._can_order_asset(security_code):
            return 0, 0
        amount = self._calculate_order_target_amount(security_code, target)
        return self.order(security_code, amount,
                          limit_price=limit_price,
                          stop_price=stop_price,
                          style=style)

    def order(self,
              security_code,
              amount,
              limit_price=None,
              stop_price=None,
              style=None):
        """
        amount 0-不交易，仅仅按最新价刷新该股票持仓总价值；负数-sell(减仓)；正数-buy(加仓)
        参考：zipline.api.order(), zipline.TradingAlgorithm()
        :return 返回提交委托交易的股票数量和价格 e.g. (1, 33.34)
        """
        if amount == 0:
            # amount=0-不交易，仅仅按最新价刷新该股票持仓总价值
            if self.position_dict.get(security_code):
                self.position_dict.get(security_code).update(0, limit_price, 0, 0, market_position=1)
            return 0, limit_price
        if not self._can_order_asset(security_code):
            return 0, 0
        try:
            self._validate_order_params(security_code, amount, limit_price, stop_price, style)
        except Exception as e:
            print("order validation error：%s" % str(e))
            return 0, 0
        amount, handled_limit_price = self._calculate_order(security_code, amount, limit_price, stop_price, style)
        if self.position_dict.get(security_code):
            self.position_dict.get(security_code).update(amount, handled_limit_price, self.min_move, self.commission, market_position=1)
        # for security_code in self.position_dict.keys():
        #     print("order....", self.position_dict[security_code].to_dict())
        return amount, handled_limit_price

    def _validate_order_params(self,
                              security_code,
                              amount,
                              limit_price,
                              stop_price,
                              style):
        if style:
            if limit_price:
                raise Exception(msg="Passing both limit_price and style is not supported.")
            if stop_price:
                raise Exception(msg="Passing both stop_price and style is not supported.")
        # TODO 其它校验：amount>0, 买受最大可交易额限制，是否持仓有足够的股票数量可以卖，下单最小股数（目前简单的以100的整数倍计算）
        if amount == 0:
            raise Exception("amount can not be equals to 0")
        if amount > 0:
            if self.position_dict.get(security_code):
                if self.position_dict.get(security_code).balance_cash < amount * limit_price:
                    raise Exception("security_code=%s, amount=%s, limit_price=%s, balance_cash=%s, buy order balance cash is not enough!!!" % (security_code, str(amount), str(limit_price), str(self.position_dict.get(security_code).balance_cash)))

    def _asymmetric_round_price(self, price, prefer_round_down, tick_size=0.01, diff=0.95):
        """
        Asymmetric rounding function for adjusting prices to the specified number
        of places in a way that "improves" the price. For limit prices, this means
        preferring to round down on buys and preferring to round up on sells.
        For stop prices, it means the reverse.
        If prefer_round_down == True:
            When .05 below to .95 above a specified decimal place, use it.
        If prefer_round_down == False:
            When .95 below to .05 above a specified decimal place, use it.

        In math-speak:
        If prefer_round_down: [<X-1>.0095, X.0195) -> round to X.01.
        If not prefer_round_down: (<X-1>.0005, X.0105] -> round to X.01.
        """
        precision = self._number_of_decimal_places(tick_size)
        multiplier = int(tick_size * (10 ** precision))
        diff -= 0.5  # shift the difference down
        diff *= (10 ** -precision)  # adjust diff to precision of tick size
        diff *= multiplier  # adjust diff to value of tick_size
        # Subtracting an epsilon from diff to enforce the open-ness of the upper
        # bound on buys and the lower bound on sells.  Using the actual system
        # epsilon doesn't quite get there, so use a slightly less epsilon-ey value.
        epsilon = float_info.epsilon * 10
        diff = diff - epsilon
        # relies on rounding half away from zero, unlike numpy's bankers' rounding
        rounded = tick_size * self._consistent_round(
            (price - (diff if prefer_round_down else -diff)) / tick_size
        )
        if self._tolerant_equals(rounded, 0.0):
            return 0.0
        return rounded

    def _calculate_order(self, security_code, amount, limit_price=None, stop_price=None, style=None):
        amount = self._round_order(amount)
        is_buy = amount > 0
        handled_limit_price = self._asymmetric_round_price(limit_price, is_buy)
        return amount, handled_limit_price

    def _round_order(self, amount):
        """
        Convert number of shares to an integer.
        By default, truncates to the integer share count that's either within
        .0001 of amount or closer to zero.
        E.g. 3.9999 -> 4.0; 5.5 -> 5.0; -5.5 -> -5.0
        """
        return int(self._round_if_near_integer(amount))

    def _round_if_near_integer(self, a, epsilon=1e-4):
        """
        Round a to the nearest integer if that integer is within an epsilon
        of a.
        """
        if abs(a - round(a)) <= epsilon:
            return round(a)
        else:
            return a

    def _consistent_round(self, val):
        if (val % 1) >= 0.5:
            return np.math.ceil(val)
        else:
            return round(val)

    def _number_of_decimal_places(self, n):
        """
        Compute the number of decimal places in a number.
        Examples
        --------
        >>> number_of_decimal_places(1)
        0
        >>> number_of_decimal_places(3.14)
        2
        >>> number_of_decimal_places('3.14')
        2
        """
        decimal = Decimal(str(n))
        return -decimal.as_tuple().exponent

    def _tolerant_equals(self, a, b, atol=10e-7, rtol=10e-7, equal_nan=False):
        """Check if a and b are equal with some tolerance.
        Parameters
        ----------
        a, b : float
            The floats to check for equality.
        atol : float, optional
            The absolute tolerance.
        rtol : float, optional
            The relative tolerance.
        equal_nan : bool, optional
            Should NaN compare equal?
        See Also
        --------
        numpy.isclose
        Notes
        -----
        This function is just a scalar version of numpy.isclose for performance.
        See the docstring of ``isclose`` for more information about ``atol`` and
        ``rtol``.
        """
        if equal_nan and np.isnan(a) and np.isnan(b):
            return True
        return math.fabs(a - b) <= (atol + rtol * math.fabs(b))

    def _calculate_order_target_amount(self, security_code, target):
        if self.position_dict.get(security_code):
            current_position = self.position_dict.get(security_code).amount
            target -= current_position
        return target

    def _can_order_asset(self, security_code):
        return self.bar_data.can_trade(security_code)

    def set_benchmark(self, benchmark_security_code):
        self._benchmark_security_code = benchmark_security_code




# coding: utf-8 or # -*- coding: utf-8 -*-
"""
如何将ELMo词向量用于中文  http://www.linzehui.me/2018/08/12/碎片知识点/如何将ELMo词向量用于中文/
通用词向量、句向量研究现状 https://www.jianshu.com/p/53de9d77e16c
代码来自：https://github.com/allenai/bilm-tf
"""


# -*- coding: utf-8 -*-
from py_common_util.common.enhanced_ordered_dict import EnhancedOrderedDict


class Position(object):
    """
    记录每只股票的持仓情况。
    init_cash_per_stock 该股票被分配的初始回测金额，e.g. 200000
    amount 剩余持有股票份数, 不可以为负数
    income 该只股票收入 e.g. (is_buy:-|is_sell:+)(amount*price) - amount*(min_move + price*commision)
    last_price 最后一次交易的价格，用于反原计算每天浮动盈亏
    market_position 当前持仓状态，多仓1，空仓-1，无仓位0
    浮动盈亏总价值=balance_cash + 浮动amount*last_price
    """
    def __init__(self, security_code, init_cash_per_stock, amount=0, start_price=0, market_position=1):
        self.security_code = security_code
        self.init_cash_per_stock = init_cash_per_stock  # 该股票分配的初始回测金额（如：20万=100万/5只股票平分）
        self.amount = amount  # 这个amount属性没有方向, 只有正数
        self.start_amount = self.amount  # 调仓周期中的起始仓位
        self.income = 0  # init_cash_per_stock + income = 剩下的现金
        self.last_price = start_price
        self.start_price = start_price  # 调仓周期中的起始价格
        self.balance_cash = self.init_cash_per_stock  # 分配给该只股票每次交易后还剩下的静态现金（不随着股价变动而浮动，已经减去了交易时的price*amount和cost）
        self.market_position = market_position
        self.order_dict = EnhancedOrderedDict()  # 记录该股票的买卖明细（不记录刷总市值的order）：Order.uuid->Order, Order:uuid, amount, price, buy_or_sell, market_position, create_date_time

    def update(self, added_amount, new_last_price, min_move, commision, market_position=1):
        """
        added_amount可以为负数, 在多仓的情况下负数表示减仓，多仓的情况下正数表示加仓,
        若added_amount=0，表示在无交易的情况下情况下，以最新的价格还原重新计算浮动盈亏的总成本价cost
        视情况而定，回测阶段系统强行自动平仓可以不计算平仓的交易成本，即传入的min_move=0,commision=0
        """
        if self.market_position > 0:
            if added_amount > 0:
                # self.income += -abs(added_amount)*new_last_price - abs(added_amount)*(min_move + new_last_price*commision)
                # self.balance_cash = self.init_cash_per_stock + self.income
                self.balance_cash += -abs(added_amount)*new_last_price - abs(added_amount)*(min_move + new_last_price*commision)
            elif added_amount == 0:  # 相当于先卖掉旧的再买回来
                # 刷新总价值时balance_cash 和 income都不变化
                # self.income += self.amount * self.last_price - self.amount * (min_move + self.last_price * commision)
                # self.income += -self.amount * new_last_price - self.amount * (min_move + new_last_price * commision)
                pass
            elif added_amount < 0:
                # self.income += abs(added_amount) * new_last_price - abs(added_amount) * (min_move + new_last_price * commision)
                # self.balance_cash = self.init_cash_per_stock + self.income
                self.balance_cash += abs(added_amount) * new_last_price - abs(added_amount) * (min_move + new_last_price * commision)
        self.amount += added_amount
        self.last_price = new_last_price
        # 第1次交易时更新start_amount和start_price
        if self.start_amount == 0:
            self.start_amount = self.amount
        if self.start_price == 0:
            self.start_price = self.last_price

    def reinit(self, init_cash_per_stock, balance_cash=0, amount=0, start_price=0, market_position=1):
        """
        重新初始化持仓(在调仓日准备重新交易旧的持仓股票的时候使用)
        clean balance，income，last_price, 只保留amount的历史值
        :return:
        """
        self.__init__(self.security_code,
                      init_cash_per_stock=init_cash_per_stock,
                      amount=amount,
                      start_price=start_price,
                      market_position=market_position)
        self.balance_cash = balance_cash

    def to_dict(self):
        """
        Creates a dictionary representing the state of this position.
        Returns a dict object of the form:
        """
        return {
            "security_code": self.security_code,
            # "start_amount": self.start_amount,
            "amount": self.amount,
            "income": self.income,
            "balance_cash": self.balance_cash,
            "start_price": self.start_price,
            "last_price": self.last_price
            # "market_position": self.market_position
        }


# -*- coding: utf-8 -*-
import tensorflow as tf
import numpy as np
from py_common_util.tensorflow.tf_utils import TFUtils
from py_common_util.common.date_utils import DateUtils
"""
TensorFlow Dataset API + Graph  https://medium.com/@ywchen88/tensorflow-dataset-api-graph-30807178cfa0
输入管道性能指南 https://tensorflow.juejin.im/performance/datasets_performance.html
TensorFlow全新的数据读取方式：Dataset API入门教程  https://zhuanlan.zhihu.com/p/30751039
tensorflow入门：tfrecord 和tf.data.TFRecordDataset https://blog.csdn.net/yeqiustu/article/details/79793454
tensorflow中的dataset http://d0evi1.com/tensorflow/datasets/
"""


class TFRecordsUtils:

    @staticmethod
    def bytes_feature(value):
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

    @staticmethod
    def int64_feature(value):
        value = value if hasattr(value, "__len__") else [value]
        return tf.train.Feature(int64_list=tf.train.Int64List(value=value))

    """
    e.g. values = [[1, 3],[2, 4],[3, 5]]
    """
    @staticmethod
    def int64_feature_list(values):
        return tf.train.FeatureList(feature=[tf.train.Feature(int64_list=tf.train.Int64List(value=values[i])) for i in range(len(values))])

    @staticmethod
    def float_feature(value):
        value = value if hasattr(value, "__len__") else [value]
        return tf.train.Feature(float_list=tf.train.FloatList(value=value))

    """
    e.g. values = [[1.1, 3.0],[-2.0, 4.1],[3.2, 5.6]]
    """
    @staticmethod
    def float_feature_list(values):
        return tf.train.FeatureList(feature=[tf.train.Feature(float_list=tf.train.FloatList(value=values[i])) for i in range(len(values))])

    """
    dict, e.g. {"feature": bytes_feature(b''), "label": int64_feature(10)}
    """
    @staticmethod
    def build_tf_example_string(dict):
        return tf.train.Example(features=tf.train.Features(feature=dict)).SerializeToString()

    @staticmethod
    def build_tf_sequence_example_string(context_dict, feature_lists_dict):
        return tf.train.SequenceExample(
            context=tf.train.Features(feature=context_dict),
            feature_lists=tf.train.FeatureLists(feature_list=feature_lists_dict)
        ).SerializeToString()

    """
    注意结束应调用writer.close()
    with get_writer(path) as writer:
        xxx
    """
    @staticmethod
    def get_writer(tfrecords_file_path):
        return tf.python_io.TFRecordWriter(tfrecords_file_path)
        # return tf.data.experimental.TFRecordWriter(tfrecords_file_path)

    @staticmethod
    def parse_dataset(tfrecords_files, parse_example_fn):
        return tf.data.TFRecordDataset(tfrecords_files).map(parse_example_fn)

    @staticmethod
    def write_summary(tfrecords_file_path, category='image', channels=1, height=0, width=0, total_count=0,
            train_count=0, valid_count=0, test_count=0, num_classes=0, data_is_normalized=False,
            using_shuffle_batch=False, remark=''):
        with TFRecordsUtils.get_writer(tfrecords_file_path) as writer:
            summary_example = tf.train.Example(features=tf.train.Features(feature={
                'feature_info': TFRecordsUtils.bytes_feature(b'0:feature_info(string),1:category(string),2:channels(int),3:height(int),4:width(int),5:total_count(int),6:train_count(int),7:valid_count(int),8:test_count(int),9:num_classes(int),10:data_is_normalized(int),11:using_shuffle_batch(int),12:remark(string)'),
                'category': TFRecordsUtils.bytes_feature(bytes(category, encoding="utf8")),
                'channels': TFRecordsUtils.int64_feature(channels),
                'height': TFRecordsUtils.int64_feature(height),
                'width': TFRecordsUtils.int64_feature(width),
                'total_count': TFRecordsUtils.int64_feature(total_count),
                'train_count': TFRecordsUtils.int64_feature(train_count),
                'valid_count': TFRecordsUtils.int64_feature(valid_count),
                'test_count': TFRecordsUtils.int64_feature(test_count),
                'num_classes': TFRecordsUtils.int64_feature(num_classes),
                'data_is_normalized': TFRecordsUtils.int64_feature(1 if data_is_normalized else 0),  # 数据是否已被正则化：true-1, false-0
                'using_shuffle_batch': TFRecordsUtils.int64_feature(1 if using_shuffle_batch else 0),  # 训练模型时是否使用shuffle batch读tfrecords的数据
                'remark': TFRecordsUtils.bytes_feature(bytes(remark, encoding="utf8"))
            }))
            writer.write(summary_example.SerializeToString())

    @staticmethod
    def serialize_data_to_string(np_data_raw=[[0.0, 0.0], [0.0, 0.0]], np_label_raw=[0, 0]):
        record_example = tf.train.SequenceExample(
            context=tf.train.Features(feature={
                "feature_info": TFRecordsUtils.bytes_feature(b'data_raw_bytes([byte]),label(int),data_raw([[float]])'),
                'data_raw_bytes': TFRecordsUtils.bytes_feature(np_data_raw.tobytes()),
                'label_raw_bytes': TFRecordsUtils.bytes_feature(np_label_raw.tobytes())
            }),
            feature_lists=tf.train.FeatureLists(feature_list={
                "data_raw": TFRecordsUtils.float_feature_list([[0.0]])
            })
        )
        return record_example.SerializeToString()

    @staticmethod
    def read_summary(tfrecords_file_path):
        """
        :param tfrecords_file_path:
        :return:
        usage:
            tuple_summary = TFRecordsUtils.read_summary("xxx.tfrecords")
            print(tuple_summary[0]) # index 0 to feature_info
        """
        def summary_parser(serialized_example):
            features = tf.compat.v1.parse_single_example(serialized_example,
                                               features={
                                                   'feature_info': tf.compat.v1.FixedLenFeature([], tf.string),
                                                   'category': tf.compat.v1.FixedLenFeature([], tf.string),
                                                   'channels': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'height': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'width': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'total_count': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'train_count': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'valid_count': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'test_count': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'num_classes': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'data_is_normalized': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'using_shuffle_batch': tf.compat.v1.FixedLenFeature([], tf.int64),
                                                   'remark': tf.compat.v1.FixedLenFeature([], tf.string)
                                               })
            return tf.cast(features['feature_info'], tf.string), \
                tf.cast(features['category'], tf.string), \
                tf.cast(features['channels'], tf.int64), \
                tf.cast(features['height'], tf.int64), \
                tf.cast(features['width'], tf.int64), \
                tf.cast(features['total_count'], tf.int64), \
                tf.cast(features['train_count'], tf.int64), \
                tf.cast(features['valid_count'], tf.int64), \
                tf.cast(features['test_count'], tf.int64), \
                tf.cast(features['num_classes'], tf.int64), \
                tf.cast(features['data_is_normalized'], tf.int64), \
                tf.cast(features['using_shuffle_batch'], tf.int64), \
                tf.cast(features['remark'], tf.string)
        dataset = tf.data.TFRecordDataset(tfrecords_file_path).map(summary_parser)
        iter_summary = tf.compat.v1.data.make_one_shot_iterator(dataset)
        batch = iter_summary.get_next()
        print("====", batch.numpy())
        sess = tf.compat.v1.Session()
        try:
            feature_info, \
            category, \
            channels, \
            height, \
            width, \
            total_count, \
            train_count, \
            valid_count, \
            test_count, \
            num_classes, \
            is_normalized, \
            using_shuffle_batch, \
            remark \
                = sess.run(batch)
            return bytes.decode(feature_info), \
                   bytes.decode(category), \
                   channels, \
                   height, \
                   width, \
                   total_count, \
                   train_count, \
                   valid_count, \
                   test_count, \
                   num_classes, \
                   True if is_normalized > 0 else False, \
                   True if using_shuffle_batch > 0 else False, \
                   bytes.decode(remark, 'utf-8')
        except tf.errors.OutOfRangeError:
            print('error occurred while reading summary records!')
        finally:
            sess.close()
        return '', 'image', 1, 1, 1, 1, 1, 1, 1, 1, False, False, ''

    @staticmethod
    def read_and_decode(tfrecords_file_path,
                        num_epoch=1,
                        category='image',
                        batch_size=128,
                        shape_data_channels_height_width=[None],
                        shape_label=[1],
                        data_is_normalized=False,
                        using_shuffle_batch=True,
                        allow_smaller_final_batch=False,
                        using_unbatch=False,
                        using_padded_batch_with_shape=(),
                        data_parser=None):
        """
        数据类型的不一致会引起下面的异常：
        tensorflow.python.framework.errors_impl.InvalidArgumentError: Input to reshape is a tensor with 1568 values, but the requested shape has 784
        处理办法：train_images = np.float32(train_images)
        :param using_padded_batch_with_shape: e.g. x,y的shape为([6],[1])
        """
        def _data_parser(serialized_example):
            contexts, features = tf.parse_single_sequence_example(serialized_example,
                                                                  context_features={
                                                                      "feature_info": tf.compat.v1.FixedLenFeature([], tf.string),
                                                                      'data_raw_bytes': tf.compat.v1.FixedLenFeature([], dtype=tf.string, default_value=''),
                                                                      'label_raw_bytes': tf.compat.v1.FixedLenFeature([], dtype=tf.string, default_value='')
                                                                  }, sequence_features={'data_raw': tf.FixedLenSequenceFeature([], dtype=tf.float32, allow_missing=True)})
            data_raw = tf.decode_raw(contexts['data_raw_bytes'], tf.float32)
            data_raw = tf.reshape(data_raw, shape_data_channels_height_width)
            label_raw = tf.decode_raw(contexts['label_raw_bytes'], tf.int64)
            label_raw = tf.squeeze(tf.reshape(label_raw, shape_label))  # 用tf.squeeze()删除大小是1的维度，e.g. [100, 1]->[100]
            if not data_is_normalized:
                data_raw = TFUtils.l2_normalization(data_raw)
            if category == 'image':
                # Display the training data_raw in the TensorBoard visualizer.
                image_data = tf.expand_dims(data_raw, 0)  # add batch_size dim
                if len(image_data.shape) < 4:
                    image_data = tf.expand_dims(image_data, 3)  # add channels dim
                # tf.summary.image('data_raw', image_data)  #[batch_size, height, width, channels]： 该句会阻塞后面的sess.run()，如需使用图片的summary功能则应该把这句提到外面
            # label_raw = tf.one_hot(label_raw, 10, on_value=1.0, off_value=0.0, axis=-1)
            return data_raw, label_raw
        drop_remainder = not allow_smaller_final_batch
        dataset = tf.data.TFRecordDataset(tfrecords_file_path).map(data_parser if data_parser else _data_parser)
        if using_unbatch:
            # 变长序列问题，The issue here is that we want to batch the data: https://stackoverflow.com/questions/49531286/tensorflow-tf-data-dataset-cannot-batch-tensors-with-different-shapes-in-compo
            dataset = dataset.apply(tf.contrib.data.unbatch())
        if len(using_padded_batch_with_shape) > 0:
            if using_shuffle_batch:
                dataset = dataset.repeat(num_epoch).padded_batch(batch_size, padded_shapes=using_padded_batch_with_shape, drop_remainder=drop_remainder).shuffle(buffer_size=1000).prefetch(buffer_size=batch_size)
            else:
                dataset = dataset.repeat(num_epoch).padded_batch(batch_size, padded_shapes=using_padded_batch_with_shape, drop_remainder=drop_remainder).prefetch(buffer_size=batch_size)
        else:
            if using_shuffle_batch:
                dataset = dataset.repeat(num_epoch).batch(batch_size, drop_remainder=drop_remainder).shuffle(buffer_size=1000).prefetch(buffer_size=batch_size)
            else:
                dataset = dataset.repeat(num_epoch).batch(batch_size, drop_remainder=drop_remainder).prefetch(buffer_size=batch_size)
        iter_data_labels = tf.compat.v1.data.make_one_shot_iterator(dataset)
        # _data_raw, _data_labels = iter_data_labels.get_next()
        #return _data_raw, _data_labels
        return iter_data_labels

    @staticmethod
    def load_np_data(tfrecords_file_path,
                     num_epoch=1,
                     num_total=0,
                     batch_size=128,
                     allow_smaller_final_batch=False,
                     category='image',
                     shape_data_channels_height_width=[],
                     shape_label=[1],
                     data_is_normalized=True,
                     using_shuffle_batch=True,
                     dense_labels_to_one_hot_classes=0,
                     lazy_load_data=False,
                     using_unbatch=False,
                     using_padded_batch_with_shape=(),
                     data_parser=None):
        """
        :param tfrecords_file_path:
        :param num_epoch:
        :param num_total:
        :param batch_size:
        :param allow_smaller_final_batch:
        :param category:
        :param shape_data_channels_height_width:
        :param data_is_normalized:
        :param using_shuffle_batch:
        :param dense_labels_to_one_hot: 0-不用处理，>0就处理labels数据为TFUtils.dense_to_one_hot(label, dense_labels_to_one_hot_classes)
        :param lazy_load_data: true 仅返回iterator, false返回数据numpy数组
        :param using_padded_batch_with_shape: e.g. x,y的shape为([6],[1])
        :return:
        usage:
        iter_data_labels, np_data_labels = \
            TFRecordsUtils.np_load_data(tfrecords_path + "/valid.tfrecords",
                                        num_total=tuple_summary[7],
                                        batch_size=128,
                                        allow_smaller_final_batch=False,
                                        category=category,
                                        shape_data_channels_height_width=[1, height, width],
                                        shape_label=[1],
                                        data_is_normalized=True,
                                        using_shuffle_batch=True,
                                        dense_labels_to_one_hot_classes=0)
        ...
        iter_data_batch = iter_data_labels.get_next()
        ...for....
            if lazy_load_data:
                ...
                batch_x, batch_y = sess.run(iter_data_batch)
            else:
                batch_x, batch_y = np_data_labels[batch_index]
        """
        start_time_int = DateUtils.now_to_int()
        print("begin read_and_decode %s, lazy_load_data=%s ..." % (tfrecords_file_path, str(lazy_load_data)))
        if not lazy_load_data:
            num_epoch = 1
        iter_data_labels = TFRecordsUtils.read_and_decode(tfrecords_file_path=tfrecords_file_path,
                                                        num_epoch=num_epoch,
                                                        category=category,
                                                        batch_size=batch_size,
                                                        shape_data_channels_height_width=shape_data_channels_height_width,
                                                        shape_label=shape_label,
                                                        data_is_normalized=data_is_normalized,
                                                        using_shuffle_batch=using_shuffle_batch,
                                                        allow_smaller_final_batch=allow_smaller_final_batch,
                                                        using_unbatch=using_unbatch,
                                                        using_padded_batch_with_shape=using_padded_batch_with_shape,
                                                        data_parser=data_parser)
        data_labels = []
        total_num_batch = TFUtils.total_num_batch(num_total, batch_size, allow_smaller_final_batch)
        if lazy_load_data:
            duration = DateUtils.calc_duration_seconds(start_time_int, DateUtils.now_to_int())
            print("finished read_and_decode on lazy_load_data=%s mode, duration(seconds)=%d ..." % (lazy_load_data, duration))
            return iter_data_labels, data_labels
        sess = tf.compat.v1.Session()
        try:
            data_raw, labels = iter_data_labels.get_next()
            for i_batch in range(total_num_batch):
                _data, _labels = sess.run([data_raw, labels])
                data_labels.append((np.array(_data, dtype=np.float32),
                                    np.array(_labels, dtype=np.int64) if dense_labels_to_one_hot_classes<1
                                        else TFUtils.dense_to_one_hot(np.asarray(_labels, dtype=np.int64),
                                    dense_labels_to_one_hot_classes)))
        except:
            print('error occurred on reading tfrecods data!')
            import traceback
            traceback.print_exc()
        finally:
            sess.close()
        duration = DateUtils.calc_duration_seconds(start_time_int, DateUtils.to_int(DateUtils.now()))
        print("finished read_and_decode on lazy_load_data=%s mode, duration(seconds)=%d ..." % (lazy_load_data, duration))
        return iter_data_labels, data_labels


if __name__ == '__main__':
    # test_path = "/tony/pycharm_projects/deep_trading/_test"
    # iterator = tf.data.Dataset.range(1000).shuffle(100).make_one_shot_iterator()
    #
    # saveable_obj = tf.contrib.data.make_saveable_from_iterator(iterator)
    # tf.add_to_collection(tf.GraphKeys.SAVEABLE_OBJECTS, saveable_obj)
    # sess = tf.compat.v1.Session()
    # init_op = tf.group(tf.global_variables_initializer(), tf.local_variables_initializer())
    # sess.run(init_op)
    # saver = tf.train.Saver()
    ####
    # for step in range(10):
    #     print('{}: {}'.format(step, sess.run(iterator.get_next())))
    #     if step == 5:
    #         # saver.save(sess, test_path + '/foo', global_step=step)
    #         saver.save(sess,
    #                    test_path + "/model.ckpt",
    #                    global_step=step,
    #                    latest_filename="model.ckpt")
    #######
    # print("*****")
    # checkpoint = tf.train.get_checkpoint_state(
    #     test_path,
    #     latest_filename="model.ckpt")
    # saver.restore(sess, checkpoint.model_checkpoint_path)
    # for step in range(6, 10):
    #     print('{}: {}'.format(step, sess.run(iterator.get_next())))
    #########################
    tfrecords_path = "/tony/pycharm_projects/deep_trading/dataset/tf_records/kcws"
    tuple_summary = TFRecordsUtils.read_summary(tfrecords_path + "/summary.tfrecords")
    category = tuple_summary[1]
    height = tuple_summary[3]
    width = tuple_summary[4]
    num_train_total = tuple_summary[6]
    num_classes = tuple_summary[9]
    dense_labels_to_one_hot_classes = 0
    lazy_load_data = True
    num_epoch = 2
    iter_data_labels, train_data_labels \
        = TFRecordsUtils.load_np_data(tfrecords_path + "/train.tfrecords",
                                    num_epoch=num_epoch,
                                    num_total=num_train_total,
                                    batch_size=100,
                                    allow_smaller_final_batch=False,
                                    category=category,
                                    shape_data_channels_height_width=[height, width],
                                    shape_label=[height],
                                    data_is_normalized=True,
                                    using_shuffle_batch=False,
                                    dense_labels_to_one_hot_classes=dense_labels_to_one_hot_classes,
                                    lazy_load_data=lazy_load_data)
    if lazy_load_data:
        total_num_batch = TFUtils.total_num_batch(num_train_total, 100, allow_smaller_final_batch=False)
        data_raw, labels = iter_data_labels.get_next()
        sess = tf.compat.v1.Session()
        data_labels = []
        for i_batch in range(total_num_batch):
            _data, _labels = sess.run([data_raw, labels])
            data_labels.append((np.array(_data, dtype=np.float32),
                                np.array(_labels, dtype=np.int64) if dense_labels_to_one_hot_classes < 1
                                else TFUtils.dense_to_one_hot(np.asarray(_labels, dtype=np.int64),
                                                              dense_labels_to_one_hot_classes)))
        train_data_labels = data_labels
    for batch_idx, (data, label) in enumerate(train_data_labels):
        print(batch_idx, len(data), len(label))




# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

torch.manual_seed(1)

######################################################################

word_to_ix = {"hello": 0, "world": 1}
embeds = nn.Embedding(2, 5)  # 2 words in vocab, 5 dimensional embeddings
lookup_tensor = torch.tensor([word_to_ix["hello"]], dtype=torch.long)
hello_embed = embeds(lookup_tensor)
print(hello_embed)


######################################################################
# An Example: N-Gram Language Modeling
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Recall that in an n-gram language model, given a sequence of words
# :math:`w`, we want to compute
#
# .. math::  P(w_i | w_{i-1}, w_{i-2}, \dots, w_{i-n+1} )
#
# Where :math:`w_i` is the ith word of the sequence.
#
# In this example, we will compute the loss function on some training
# examples and update the parameters with backpropagation.
#

CONTEXT_SIZE = 2
EMBEDDING_DIM = 10
# We will use Shakespeare Sonnet 2
test_sentence = """When forty winters shall besiege thy brow,
And dig deep trenches in thy beauty's field,
Thy youth's proud livery so gazed on now,
Will be a totter'd weed of small worth held:
Then being asked, where all thy beauty lies,
Where all the treasure of thy lusty days;
To say, within thine own deep sunken eyes,
Were an all-eating shame, and thriftless praise.
How much more praise deserv'd thy beauty's use,
If thou couldst answer 'This fair child of mine
Shall sum my count, and make my old excuse,'
Proving his beauty by succession thine!
This were to be new made when thou art old,
And see thy blood warm when thou feel'st it cold.""".split()
# we should tokenize the input, but we will ignore that for now
# build a list of tuples.  Each tuple is ([ word_i-2, word_i-1 ], target word)
trigrams = [([test_sentence[i], test_sentence[i + 1]], test_sentence[i + 2])
            for i in range(len(test_sentence) - 2)]
# print the first 3, just so you can see what they look like
print(trigrams[:3])

vocab = set(test_sentence)
word_to_ix = {word: i for i, word in enumerate(vocab)}


class NGramLanguageModeler(nn.Module):

    def __init__(self, vocab_size, embedding_dim, context_size):
        super(NGramLanguageModeler, self).__init__()
        self.embeddings = nn.Embedding(vocab_size, embedding_dim)
        self.linear1 = nn.Linear(context_size * embedding_dim, 128)
        self.linear2 = nn.Linear(128, vocab_size)

    def forward(self, inputs):
        embeds = self.embeddings(inputs).view((1, -1))
        out = F.relu(self.linear1(embeds))
        out = self.linear2(out)
        log_probs = F.log_softmax(out, dim=1)
        return log_probs


losses = []
loss_function = nn.NLLLoss()
model = NGramLanguageModeler(len(vocab), EMBEDDING_DIM, CONTEXT_SIZE)
optimizer = optim.SGD(model.parameters(), lr=0.001)

for epoch in range(10):
    total_loss = 0
    for context, target in trigrams:

        # Step 1. Prepare the inputs to be passed to the model (i.e, turn the words
        # into integer indices and wrap them in tensors)
        context_idxs = torch.tensor([word_to_ix[w] for w in context], dtype=torch.long)

        # Step 2. Recall that torch *accumulates* gradients. Before passing in a
        # new instance, you need to zero out the gradients from the old
        # instance
        model.zero_grad()

        # Step 3. Run the forward pass, getting log probabilities over next
        # words
        log_probs = model(context_idxs)

        # Step 4. Compute your loss function. (Again, Torch wants the target
        # word wrapped in a tensor)
        loss = loss_function(log_probs, torch.tensor([word_to_ix[target]], dtype=torch.long))

        # Step 5. Do the backward pass and update the gradient
        loss.backward()
        optimizer.step()

        # Get the Python number from a 1-element Tensor by calling tensor.item()
        total_loss += loss.item()
    losses.append(total_loss)
print(losses)  # The loss decreased every iteration over the training data!


######################################################################
# Exercise: Computing Word Embeddings: Continuous Bag-of-Words
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The Continuous Bag-of-Words model (CBOW) is frequently used in NLP deep
# learning. It is a model that tries to predict words given the context of
# a few words before and a few words after the target word. This is
# distinct from language modeling, since CBOW is not sequential and does
# not have to be probabilistic. Typcially, CBOW is used to quickly train
# word embeddings, and these embeddings are used to initialize the
# embeddings of some more complicated model. Usually, this is referred to
# as *pretraining embeddings*. It almost always helps performance a couple
# of percent.
#
# The CBOW model is as follows. Given a target word :math:`w_i` and an
# :math:`N` context window on each side, :math:`w_{i-1}, \dots, w_{i-N}`
# and :math:`w_{i+1}, \dots, w_{i+N}`, referring to all context words
# collectively as :math:`C`, CBOW tries to minimize
#
# .. math::  -\log p(w_i | C) = -\log \text{Softmax}(A(\sum_{w \in C} q_w) + b)
#
# where :math:`q_w` is the embedding for word :math:`w`.
#
# Implement this model in Pytorch by filling in the class below. Some
# tips:
#
# * Think about which parameters you need to define.
# * Make sure you know what shape each operation expects. Use .view() if you need to
#   reshape.
#

CONTEXT_SIZE = 2  # 2 words to the left, 2 to the right
raw_text = """We are about to study the idea of a computational process.
Computational processes are abstract beings that inhabit computers.
As they evolve, processes manipulate other abstract things called data.
The evolution of a process is directed by a pattern of rules
called a program. People create programs to direct processes. In effect,
we conjure the spirits of the computer with our spells.""".split()

# By deriving a set from `raw_text`, we deduplicate the array
vocab = set(raw_text)
vocab_size = len(vocab)

word_to_ix = {word: i for i, word in enumerate(vocab)}
data = []
for i in range(2, len(raw_text) - 2):
    context = [raw_text[i - 2], raw_text[i - 1],
               raw_text[i + 1], raw_text[i + 2]]
    target = raw_text[i]
    data.append((context, target))
print(data[:5])


class CBOW(nn.Module):

    def __init__(self):
        pass

    def forward(self, inputs):
        pass

# create your model and train.  here are some functions to help you make
# the data ready for use by your module


def make_context_vector(context, word_to_ix):
    idxs = [word_to_ix[w] for w in context]
    return torch.tensor(idxs, dtype=torch.long)


make_context_vector(data[0][0], word_to_ix)  # example


# -*- coding:utf-8 -*-
import tushare as ts

from py_common_util.tushare import PandasUtils
from py_common_util.tushare import MySQLUtils
import datetime
import time
'''
ref: http://tushare.org/
ref: 一个提升选股技能和效果的接口 https://mp.weixin.qq.com/s/tzaV8sfhF6BNJP83cI_q2g
'''


class TushareWrapper(object):

    @property
    def db_name(self):
        return 'tushare_2017'

    @property
    def last_trading_date(self):
        """每个交易日收盘后更新trading_date, tushare在交易日17:00后更新最新数据文件"""
        """如需补抓历史(N-x, N)天的历史数据，则将last_trading_date设为最新日期，调用daily_real_data_update()后再将start_date设为N-x日期调hist_k_data_update()"""
        return '20180510'

    def __init__(self):
        print("tushare -v: " + ts.__version__)
        self.mysqlUtils = MySQLUtils()
        self.pandasUtils = PandasUtils()

    def daily_real_data_update(self, trading_date=None):
        trading_date = self.last_trading_date if trading_date is None else trading_date
        print("每日实时数据更新，dailyUpdate()..., trading_date=" + trading_date)
        print('注意：store_stock_basics和store_index应在每个交易日的23:59前就更新')
        self.store_stock_basics(trading_date)
        self.store_index(trading_date)
        ##########################################
        # self.store_today_all(trading_date)  # 有问题(20171020)
        # self.store_sina_dd(trading_date=trading_date)  # 有问题(20171019)
        # self.store_top_data(trading_date)
        # self.store_stock_classified_data(trading_date)

    def hist_k_data_update(self, special_codes=[], from_date='', end_date='', trading_date=None, autype='qfq', ktype='D', index=False, should_exclude_saved_code_as_rerun=True):
        """更新历史K线数据支持断开续取, 注意应每日备份该方法产生的数据"""
        """特别注意：该表数据巨大，操作时要特别谨慎！只有在每个新交易日第1次运行该方法的时候should_exclude_saved_code_as_rerun才为False并且设置from_date=trading_date。第2次及以后应为True"""
        """from_date/end_date, e.g. '20171013' """
        """注意：df为空的情况，原因是抓取页面的HTML结构或页面JSON有问题"""
        """注意：应该在开盘前抓取上个交易日的历史数据"""
        trading_date = self.last_trading_date if trading_date is None else trading_date
        excluded_codes = []
        if should_exclude_saved_code_as_rerun:
            saved_codes = tushare_wrapper.query_all_k_data_stock_code(autype=autype, ktype=ktype, index=index)
            excluded_codes = saved_codes
            print("as re-run, saved_codes=" + str(saved_codes) + ", len(saved_codes)=" + str(len(saved_codes)))
        print("hist_k_data_update()..., from_date=%s, end_date=%s, trading_date=%s" % (from_date, end_date, trading_date))
        self.store_k_data(special_codes=special_codes, excluded_codes=excluded_codes, from_date=from_date, end_date=end_date, autype=autype, ktype=ktype, index=index, trading_date=trading_date)

    def store_top_data(self, trading_date=None):
        """龙虎榜数据: 龙虎榜数据接口提供历史龙虎榜上榜股票数据"""
        trading_date = self.last_trading_date if trading_date is None else trading_date
        # 每日龙虎榜列表
        print('top_list...')
        top_df = ts.top_list(self.stock_date_format(trading_date))
        self.mysqlUtils.append_data(top_df, 'top_list')
        # 个股上榜统计
        print('cap_tops...')
        cap_tops_df = ts.cap_tops()
        cap_tops_df['date'] = trading_date
        self.mysqlUtils.append_data(cap_tops_df, 'cap_tops')
        # 营业部上榜统计
        print('broker_tops...')
        broker_tops_df = ts.broker_tops()
        broker_tops_df['date'] = trading_date
        self.mysqlUtils.append_data(broker_tops_df, 'broker_tops')
        # 龙虎榜机构席位追踪
        print('inst_tops...')
        inst_tops_df = ts.inst_tops()
        inst_tops_df['date'] = trading_date
        self.mysqlUtils.append_data(inst_tops_df, 'inst_tops')
        # 龙虎榜机构席位成交明细
        print('inst_detail...')
        inst_detail_df = ts.inst_detail()
        self.mysqlUtils.append_data(inst_detail_df, 'inst_detail')

    def store_stock_classified_data(self, trading_date=None):
        """股票分类数据, 需要逐个抓取入库"""
        trading_date = self.last_trading_date if trading_date is None else trading_date
        # 行业分类
        # print('get_industry_classified...')
        # industry_df = ts.get_industry_classified()  # 修改了该方法里的代码
        # industry_df['date'] = trading_date
        # self.mysqlUtils.append_data(industry_df, 'industry_classified')
        # 概念分类
        # print('get_concept_classified...')
        # concept_df = ts.get_concept_classified()  # 修改了该方法里的代码
        # concept_df['date'] = trading_date
        # self.mysqlUtils.append_data(concept_df, 'concept_classified')
        # ####以下数据可以一次性获取
        # 地域分类
        # print('get_area_classified...')
        # area_df = ts.get_area_classified()
        # area_df['date'] = trading_date
        # self.mysqlUtils.append_data(area_df, 'area_classified')
        # 中小板分类, 获取中小板股票数据，即查找所有002开头的股票
        # print('get_sme_classified...')
        # sme_df = ts.get_sme_classified()
        # sme_df['date'] = trading_date
        # self.mysqlUtils.append_data(sme_df, 'sme_classified')
        # 创业板分类: 获取创业板股票数据，即查找所有300开头的股票
        # print('get_gem_classified...')
        # gem_df = ts.get_gem_classified()
        # gem_df['date'] = trading_date
        # self.mysqlUtils.append_data(gem_df, 'gem_classified')
        # 风险警示板分类: 获取风险警示板股票数据，即查找所有st股票
        # print('get_st_classified...')
        # st_df = ts.get_st_classified()
        # st_df['date'] = trading_date
        # self.mysqlUtils.append_data(st_df, 'st_classified')
        # # 沪深300成份及权重: 获取沪深300当前成份股及所占权重
        ####以下有问题 20171021
        # print('get_hs300s')
        # hs300s_df = ts.get_hs300s()
        # hs300s_df['date'] = trading_date
        # self.mysqlUtils.append_data(hs300s_df, 'hs300s')
        # # 上证50成份股
        # print('get_sz50s...')
        # sz50s_df = ts.get_sz50s()
        # sz50s_df['date'] = trading_date
        # self.mysqlUtils.append_data(sz50s_df, 'sz50s')
        # # 中证500成份股(有问题：20171018)
        # print('get_zz500s...')
        # zz500s_df = ts.get_zz500s()
        # zz500s_df['date'] = trading_date
        # self.mysqlUtils.append_data(zz500s_df, 'zz500s')
        # 终止上市股票列表: 获取已经被终止上市的股票列表，数据从上交所获取，目前只有在上海证券交易所交易被终止的股票。
        # print('get_terminated...')
        # terminated_df = ts.get_terminated()
        # terminated_df['date'] = trading_date
        # self.mysqlUtils.append_data(terminated_df, 'terminated')
        # 暂停上市股票列表: 获取被暂停上市的股票列表，数据从上交所获取，目前只有在上海证券交易所交易被终止的股票。
        # print('get_suspended...')
        # suspended_df = ts.get_suspended()
        # suspended_df['date'] = trading_date
        # self.mysqlUtils.append_data(suspended_df, 'suspended')

    def store_stock_basics(self, trading_date=None):
        print("get_stock_basics()......trading_date=%s" % trading_date)
        df = ts.get_stock_basics(self.stock_date_format(date=trading_date))
        df['date'] = trading_date
        self.mysqlUtils.append_data(df, 'stock_basics')

    def store_k_data(self, special_codes=[], excluded_codes=[], from_date='', end_date='', autype='qfq', ktype='D', index=False, trading_date=None):
        start_time = time.time()
        print("store_k_data()..., from_date=%s, end_date=%s, trading_date=%s, ktype=%s, index=%s" % (from_date, end_date, trading_date, ktype, index))
        original_table_name = 'k_data_'
        table_name = 'k_data_'
        # D=日k线 W=周 M=月 5=5分钟 15=15分钟 30=30分钟 60=60分钟，默认为D
        if ktype == 'D':
            original_table_name = 'k_index_day' if index else 'k_data_day'
        elif ktype == 'W':
            original_table_name = 'k_index_week' if index else 'k_data_week'
        elif ktype == 'M':
            original_table_name = 'k_index_month' if index else 'k_data_month'
        elif ktype == '5':
            original_table_name = 'k_index_5min' if index else 'k_data_5min'
        elif ktype == '15':
            original_table_name = 'k_index_15min' if index else 'k_data_15min'
        elif ktype == '30':
            original_table_name = 'k_index_30min' if index else 'k_data_30min'
        elif ktype == '60':
            original_table_name = 'k_index_60min' if index else 'k_data_60min'
        if autype == 'qfq':
            original_table_name += '_qfq'
        elif autype == 'hfq':
            original_table_name += '_hfq'
        """特别注意：为防止重复保存或弄乱已有的巨大的历史数据，每次重新保存数据时都重建新表，待检查无误时再谨慎运行增量更新的SQL"""
        """
        增量更新SQL:
        INSERT INTO k_index_day_qfq (`index`, date, open, close, high, low, volume, code)  
        SELECT `index`, date, open, close, high, low, volume, code
          FROM `k_index_day_qfq_20171017`
        """
        if len(special_codes) > 0:
            table_name = original_table_name + "_" + str(special_codes) + '_' + trading_date
        else:
            table_name = original_table_name + '_' + trading_date
        print("hist k Data will be save to table_name: " + table_name)
        print("Please copy SQL and execute manumally: \n" + "INSERT INTO %s (`index`, date, open, close, high, low, volume, code)  \n" \
        "SELECT `index`, date, open, close, high, low, volume, code \n" "FROM `%s`" % (original_table_name, table_name))

        if index:
            tupes = self.query_all_index_code(trading_date)
        else:
            tupes = self.query_all_stock_code(trading_date)
        step = 0
        for code, time_to_market in tupes:
            step += 1
            if len(special_codes) > 0 and code not in special_codes:
                continue
            category_code = code
            if index:
                index_category = ''
                if code.find('000') == 0:
                    index_category = 'sh'
                elif code.find('399') == 0:
                    index_category = 'sz'
                category_code = index_category + code
            # 新股的time_to_market为0
            if (time_to_market < 1 and not index) or (category_code in excluded_codes):
                print("code: " + str(code) + " was excluded! or time_to_market(only for new stock):" + str(time_to_market) + " < 1")
                continue
            if step < 1 or ((step + 1) % 100 == 0) or ((step + 1) >= self.get_row_num(tupes)):
                print("store_k_data... step=%d, total=%d" % (step+1, self.get_row_num(tupes)))
            df = self.get_k_data(code=code,
                                 autype=autype,
                                 ktype=ktype,
                                 index=index,
                                 start=self.stock_date_format(str(time_to_market) if (from_date == '' or from_date is None) else from_date),
                                 end='' if (end_date == '' or end_date is None) else self.stock_date_format(trading_date))
            self.mysqlUtils.append_data(df, table_name)  # 表k_data_day/k_index_day已被手工重建索引(类型：text->varchar, 索引：index,code/code)，删表结构时需要谨慎
            print("saved_codes(should be exclude as re-run):" + str(code) + ", df row num: " + str(self.get_row_num(df)))
        print("store_k_data is ok...total take time(seconds): %d" % (time.time() - start_time))

    def query_all_stock_code(self, trading_date=None):
        df = self.pandasUtils.read_sql("SELECT code, timeToMarket FROM " + self.db_name + ".stock_basics where date=" + trading_date)
        return df.values

    def query_all_index_code(self, trading_date=None):
        df = self.pandasUtils.read_sql("SELECT code, 0 FROM " + self.db_name + ".index_data where date=" + trading_date)
        return df.values

    def query_stock_time_to_market(self, code):
        """取上市时间作为历史数据的start date(e.g. 19900101), 0代表新股尚未上市的股票"""
        df = self.pandasUtils.read_sql("SELECT timeToMarket FROM " + self.db_name + ".stock_basics where code='" + code + "'")
        return df.timeToMarket.values

    def query_all_k_data_stock_code(self, autype='qfq', ktype='D', index=False):
        """获取已经保存到K线表里的所有股票code或指数code"""
        print("query_all_k_data_stock_code..., autype=%s, ktype=%s, index=%s" % (autype, ktype, index))
        table_name = ''
        if ktype == 'D':
            table_name = 'k_index_day' if index else 'k_data_day'
        elif ktype == 'W':
            table_name = 'k_index_week' if index else 'k_data_week'
        elif ktype == 'M':
            table_name = 'k_index_month' if index else 'k_data_month'
        elif ktype == '5':
            table_name = 'k_index_5min' if index else 'k_data_5min'
        elif ktype == '15':
            table_name = 'k_index_15min' if index else 'k_data_15min'
        elif ktype == '30':
            table_name = 'k_index_30min' if index else 'k_data_30min'
        elif ktype == '60':
            table_name = 'k_index_60min' if index else 'k_data_60min'
        if autype == 'qfq':
            table_name += '_qfq'
        elif autype == 'hfq':
            table_name += '_hfq'
        df = self.pandasUtils.read_sql("SELECT distinct(code) FROM " + self.db_name + "." + table_name)
        return df.values.flatten()  # e.g. [['000005'],['000006']] -> ['000005', '000006']

    def get_k_data(self, code=None, autype='qfq', start='', end='', ktype='D', index=False):
        """ 本接口只能获取近3年的日线index数据，适合搭配均线数据进行选股和分析，如果需要全部历史数据，请调用接口get_h_data() """
        # print("get_k_data()......,code=%s, start=%s, end=%s, ktype=%s, index=%s" % (code, start, end, ktype, index))
        df = ts.get_k_data(code=code, autype=autype, start=start, end=end, ktype=ktype, index=index, retry_count=10)
        return df

    def get_h_data(self, code=None, from_date=None, end_date=None, index=False, autype='qfq', drop_factor=False):
        """获取所有复权的历史数据, 取前复权的数据"""
        """start/end, e.g. '20171017'"""
        print("get_h_data()......")
        df = ts.get_h_data(code=code,
                           start=self.stock_date_format(from_date),
                           end=self.stock_date_format(end_date),
                           index=index,
                           autype=autype,
                           drop_factor=drop_factor)
        return df

    def store_today_all(self, date=None):
        """一次性获取最近一个日交易日所有股票的交易数据"""
        """ 数据每日盘中实时更新, 应该盘后抓取当天的最新数据 """
        print("store_today_all()......")
        df = ts.get_today_all()
        df['date'] = self.now_date_str() if date is None else date
        self.mysqlUtils.append_data(df, 'today_all')

    def store_index(self, trading_date=None):
        """获取大盘指数实时行情列表，以表格的形式展示大盘指数实时行情, 收盘后就是日K历史的行情"""
        print("store_index()......")
        trading_date = self.last_trading_date if trading_date is None else trading_date
        df = ts.get_index()
        df['date'] = trading_date
        self.mysqlUtils.append_data(df, 'index_data')

    def store_sina_dd(self, vol=400, trading_date=None):
        """获取大单交易数据，默认为大于等于400手，数据来源于新浪财经"""
        print("store_sina_dd()......trading_date=" + trading_date)
        date = self.stock_date_format(date=trading_date)
        step = 0
        tupes = self.query_all_stock_code(trading_date)
        for code, _ in tupes:
            df = ts.get_sina_dd(code=code, date=date, vol=vol)
            if df is not None:
                step += 1
                if step < 1 or ((step + 1) % 100 == 0) or ((step + 1) >= self.get_row_num(tupes)):
                    print("step=%d, total=%d" % (step+1, self.get_row_num(tupes)))
                df['date'] = trading_date
                self.mysqlUtils.append_data(df, 'sina_dd')

    def stock_date_format(self, date='19900101', from_format='%Y%m%d', to_format='%Y-%m-%d'):
        """e.g. 19900101->1990-01-01"""
        if date == '0' or date == '':
            return ''
        if date is None:
            return None
        return datetime.datetime.strptime(date, from_format).strftime(to_format)

    def now_date_str(self):
        return datetime.datetime.now().strftime('%Y%m%d')

    def get_row_num(self, df=None):
        if df is None:
            return 0
        return df.shape[0]


if __name__ == '__main__':
    tushare_wrapper = TushareWrapper()
    print("(18:00->24:00, 00:00->9:00)请尽量在每个交易日的盘后及第2天的开盘前抓取行情数据。。。。。。")
    # tushare_wrapper.daily_real_data_update()  # 增量每日实时数据更新
    # 增量每日更新(第1次更新should_exclude_saved_code_as_rerun=False)
    # tushare_wrapper.hist_k_data_update(index=True,
    #                                    # special_codes=['000001'],
    #                                    from_date=tushare_wrapper.last_trading_date, # ''为历史最早的交易日
    #                                    end_date=tushare_wrapper.last_trading_date,  # ''为最近交易日
    #                                    autype='qfq',
    #                                    ktype='D',
    #                                    should_exclude_saved_code_as_rerun=False)
    # tushare_wrapper.hist_k_data_update(index=False,
    #                                    from_date=tushare_wrapper.last_trading_date,  # ''为历史最早的交易日
    #                                    end_date=tushare_wrapper.last_trading_date,   # ''为最近交易日
    #                                    autype='qfq',
    #                                    ktype='D',
    #                                    should_exclude_saved_code_as_rerun=False)  # 全量从第1行开始insert





# -*- coding: utf-8 -*-


class BacktestSetting(object):
    """回测的基本设置"""
    def __init__(self, trade_mode, start_date, end_date, benchmark_code, init_amount:int=1000000, cost:float=0.001):
        self.trade_mode_type = 0  # 目前只有0:定期轮动，回测类型 0:定期轮动，下标从0开始
        self.trade_mode = trade_mode  # 交易模型：0: 定期轮动
        self.start_date = start_date[0:4]+"-"+start_date[4:6]+"-"+start_date[6:8]  # 起始日期 e.g. "2019-01-01"
        self.end_date = end_date[0:4]+"-"+end_date[4:6]+"-"+end_date[6:8]  # 结束日期 e.g. "2019-01-01"
        self.init_amount = int(init_amount)  # 初始金额(元)
        self.benchmark_code = benchmark_code  # 指数作为基准标的
        self.cost = float(cost)  # 手续费  0:零, 0.001: 千分之一, 0.002:千分之2, 0.005:千分之五, 0.01:千分之十




# -*- coding: utf-8 -*-
import tensorflow as tf
import numpy as np
# 解决Warning：Your CPU supports instructions that this TensorFlow binary was not compiled to use: AVX2 FMA
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
# 当你和其他人共用同台机器上的多卡gpu时，防止和别人争用资源，每个人都最好指定使用哪张GPU卡
# os.environ["CUDA_VISIBLE_DEVICES"] = "0,1"


class TFUtils:

    @staticmethod
    def empty_zero_data(shape_channels_height_width=[1, 28, 28], dtype=np.float32):
        """
        返回length=0的数据
        :param shape_channels_height_width:
        e.g. [(np.zeros((0, 1, 28, 28), dtype=np.float32), np.zeros((0, 10), dtype=np.int64))]
        :return:
        """
        return np.zeros((0,
                        shape_channels_height_width[0],
                        shape_channels_height_width[1],
                        shape_channels_height_width[2]), dtype=dtype)

    @staticmethod
    def total_num_batch(num_total, batch_size, allow_smaller_final_batch=False):
        num_batch = num_total // batch_size
        if allow_smaller_final_batch:
            if num_total > (num_batch * batch_size):
                return num_batch + 1
        else:
            # 舍去最后不能够成1批的剩余训练数据部分
            return num_batch

    @staticmethod
    def batchify_np_data_label(tuple_np_data_label, batch_size, allow_smaller_final_batch=False):
        """
        :param tuple_np_data_label:
        :param batch_size:
        :return:
        使用：
        ......
        batched_data_label = TFUtils.batchify_np_data_label(np_test_data, parser_args.eval_batch_size)
        for batch_idx, (data, label) in enumerate(batched_data_label):
            data, label = ......
        ......
        """
        if tuple_np_data_label is None or len(tuple_np_data_label) < 1:
            return []
        data = tuple_np_data_label[0][0]
        label = tuple_np_data_label[0][1]
        num_batch = TFUtils.total_num_batch(data.shape[0], batch_size, allow_smaller_final_batch)
        data = data[:num_batch*batch_size]
        label = label[:num_batch*batch_size]
        batched_data = np.split(data, num_batch)
        batched_label = np.split(label, num_batch)
        return list(zip(batched_data, batched_label))

    @staticmethod
    def l2_normalization(x):
        axis = TFUtils.get_axis(x)
        return tf.nn.l2_normalize(x, axis=axis)

    @staticmethod
    def std_normalization(x):
        axis = TFUtils.get_axis(x)
        batch_mean, batch_var = tf.nn.moments(x=x, axes=axis, name='moments')
        return tf.divide(tf.subtract(x, batch_mean), tf.sqrt(batch_var + TFUtils.get_small_epsilon()))

    @staticmethod
    def get_small_epsilon():
        return 1e-3

    @staticmethod
    def get_axis(x):
        return list(range(len(x.shape) - 1))

    @staticmethod
    def read_mnist_from_local(data_dir="../../dataset/MNIST_data/"):
        """e.g. data_dir=sys.path[0] + '/../../dataset/MNIST_data/'"""
        import tensorflow.examples.tutorials.mnist.input_data as mnist_input_data
        mnist = mnist_input_data.read_data_sets(data_dir, one_hot=True)
        print("MNIST loaded!")
        return mnist

    @staticmethod
    def tf_l2_normal_weight_biases(n_hidden, n_classes, num_hidden=1):
        weights = tf.Variable(tf.nn.l2_normalize(tf.random.normal([num_hidden*n_hidden, n_classes]), [0, 1]))
        biases = tf.Variable(tf.random.normal([n_classes]))
        return weights, biases

    @staticmethod
    def tf_random_normal_weight_biases(n_hidden, n_classes, num_hidden=1):
        weights = tf.Variable(tf.random.normal([num_hidden*n_hidden, n_classes]))
        biases = tf.Variable(tf.random.normal([n_classes]))
        return weights, biases

    @staticmethod
    def tf_truncate_normal_weight_biases(n_hidden, n_classes, num_hidden=1):
        weights = tf.Variable(tf.random.truncated_normal([num_hidden*n_hidden, n_classes], stddev=.075, seed=1234, name='final_weights'))
        biases = tf.Variable(tf.random.normal([n_classes]))
        return weights, biases

    @staticmethod
    def dense_to_one_hot(labels_dense, num_classes=10):
        """
        Convert class labels from scalars to one-hot vectors.
        labels_dense e.g. np.asarray([5, 0, 3, 2, 4, 9, 6, 7, 2, 1, 0, 0, 1, 4, 5, 2, 5, 6, 7, 8])
        as example: dense_to_one_hot(np.asarray([8]), 10)
        ref to: tensorflow/contrib/learn/python/learn/datasets/mnist.py#dense_to_one_hot()
        """
        num_labels = labels_dense.shape[0]
        index_offset = np.arange(num_labels) * num_classes
        labels_one_hot = np.zeros((num_labels, num_classes))
        labels_one_hot.flat[index_offset + labels_dense.ravel()] = 1
        return labels_one_hot

    @staticmethod
    def one_hot_to_dense(labels_dense_one_hot=[]):
        """
        for example:
        one_hot_to_dense([ 0.  1.  0.  0.  0.  0.  0.  0.  0.  0.])
        result index is: [1], and the result shuld be: index+1
        one_hot_to_dense([ 0.  0.  0.  1.  0.  0.  0.  0.  0.  0.])
        result index is: [3], and the result shuld be: index+1
        """
        if np.asarray(labels_dense_one_hot.shape).size <= 1:
            return labels_dense_one_hot
        with tf.compat.v1.Session() as sess:
            return sess.run(tf.argmax(input=labels_dense_one_hot, axis=1))

    @staticmethod
    def average(pixel):
        # #ref to :https://samarthbhargav.wordpress.com/2014/05/05/image-processing-with-python-rgb-to-grayscale-conversion/
        # #1.convert RGB image to gray image
        # return (pixel[0] + pixel[1] + pixel[2]) / 3
        # #2.using numpy average
        # #3.G = R*0.299 + G*0.587 + B*0.114
        return 0.299 * pixel[0] + 0.587 * pixel[1] + 0.114 * pixel[2]

    @staticmethod
    def np_to_tensor(np_data):
        return tf.convert_to_tensor(value=np_data)

    @staticmethod
    def check_available_gpus(show_info=False):
        from tensorflow.python.client import device_lib
        local_devices = device_lib.list_local_devices()
        gpu_names = [x.name for x in local_devices if x.device_type == 'GPU']
        gpu_num = len(gpu_names)
        if show_info:
            print('{0} GPUs are detected : {1}'.format(gpu_num, gpu_names))
        return gpu_num

    @staticmethod
    def cuda_is_available(show_info=False):
        return TFUtils.check_available_gpus(show_info) > 0

    @staticmethod
    def device(device="gpu", device_index=0):
        """
        :param: device="gpu" 或 "cpu"
        :param: device_index
        :return:
        usage:
        ......
        with TFUtils.device("gpu", 1):
            sum_operation = tf.reduce_sum(dot_operation)
            ......
        """
        if TFUtils.cuda_is_available(show_info=False) and device == 'gpu':
            return tf.device(tf.compat.v1.DeviceSpec(device_type="GPU", device_index=device_index))
        else:
            return tf.device(tf.compat.v1.DeviceSpec(device_type="CPU", device_index=device_index))

    @staticmethod
    def tf_get_length3(sequence):
        """
        去掉batch padded功能，tensorflow特有的get_length，可用于nlp变长序列的bilstm，pytorch没有对应方法
        参考：Variable Sequence Lengths in TensorFlow  https://danijar.com/variable-sequence-lengths-in-tensorflow/
        :param sequence:
        :return:
        e.g.
        [
            [[1,2,3],[1,0,0]],
            [[1,2,3],[4,5,6]]
        ] -> [2,2]
        [
            [[1,2,3],[0,0,0]],
            [[1,2,3],[4,5,6]]
        ] -> [1,2]
        """
        used = tf.sign(tf.reduce_max(input_tensor=tf.abs(sequence), axis=2))
        length = tf.reduce_sum(input_tensor=used, axis=1)
        length = tf.cast(length, tf.int32)
        return length

    @staticmethod
    def tf_get_length2(sequence):
        """
        去掉batch padded功能，tensorflow特有的get_length，可用于nlp变长序列的bilstm, pytorch没有对应方法
        参考：Variable Sequence Lengths in TensorFlow  https://danijar.com/variable-sequence-lengths-in-tensorflow/
        :param x:
        :return:
        e.g. [[1,2,3],[4,5,6]] -> [3,3]
        [[0,2,3],[4,5,6]] -> [2,3]
        """
        used = tf.sign(tf.abs(sequence))  # for x的维度为：[100,80]的场景
        length = tf.reduce_sum(input_tensor=used, axis=1)
        length = tf.cast(length, tf.int32)
        return length

    @staticmethod
    def tf_get_length1(sequence):
        """
        去掉batch padded功能，tensorflow特有的get_length，可用于nlp变长序列的bilstm, pytorch没有对应方法
        参考：Variable Sequence Lengths in TensorFlow  https://danijar.com/variable-sequence-lengths-in-tensorflow/
        :param x:
        :return:
        e.g. [1,2,8] -> 3
        [1,0,8] -> 2
        [0,0,0] -> 0
        [-4,-2,-3] -> 3
        [-4,0,-3] -> 2
        """
        used = tf.sign(tf.abs(sequence))  # for x的维度为：[100]的场景
        length = tf.reduce_sum(input_tensor=used, axis=0, keepdims=False)
        length = tf.cast(length, tf.int32)
        return length


if __name__ == '__main__':
    from py_common_util.common.common_utils import CommonUtils
    iter = CommonUtils.cycle()
    print(next(iter))
    print(next(iter))
    print(next(iter))
    print(next(iter))
    print(next(iter))
    print(next(iter))



# coding: utf-8 or # -*- coding: utf-8 -*-
from sqlalchemy import create_engine
"""
解决MYSQL更新数据Error Code: 1175. You are using safe update
SET SQL_SAFE_UPDATES = 0; 降低数据库修改模式
SET SQL_SAFE_UPDATES = 1; 提高数据库修改模式
"""


class MySQLUtils(object):

    @property
    def engine(self):
        return self._engine

    def __init__(self):
        self._engine = create_engine('mysql+pymysql://root:123456@127.0.0.1:3306/tushare_2017?charset=utf8')

    def append_data(self, df, table_name):
        # 追加数据到现有表
        df.to_sql(table_name, self.engine, if_exists='append')

    def replace_data(self, df, table_name):
        # 覆盖数据到现有表
        df.to_sql(table_name, self.engine, if_exists='replace')

    def insert_data_safely(self, df, table_name):
        # 如果表存在就报错
        df.to_sql(table_name, self.engine, if_exists='fail')


if __name__ == '__main__':
    mysqlUtil = MySQLUtils()
    import pandas as pd
    df = pd.read_sql("SELECT * FROM tushare_2017.stock_basics", mysqlUtil.engine)
    print(df)



# -*- coding:utf-8 -*-
import talib
from py_common_util.tushare import PandasUtils
from py_common_util.tushare import MySQLUtils
'''
量化工具TA-Lib 使用例子 http://30daydo.com/article/196
'''


class TalibUtils(object):

    @property
    def version(self):
        return 'tushare_2017'

    def __init__(self):
        print("TA-lib -v: " + talib.__version__)


if __name__ == '__main__':
    talib_util = TalibUtils()
    # talib_util.MA(df.close.values, timeperiod=5, matype=0)


# -*- coding: utf-8 -*-


class TradeContext(object):
    """交易过程的上下文，保存中间变量值, 参考zipline.algorithm.TradingAlgorithm"""
    def __init__(self, *args, **kwargs):
        self.trade_calendar_list = []  # 交易日历执行计划
        self.current_bar_item_day_date = None  # K线当日的日期，如"2019-08-01"
        self.current_bar_item_kline_date = None  # 当前k线的日期，分钟K线如"2019-08-01 15:30" 日K如"2019-08-01"
        self.stock_type = ""  # 股票类型 HK: 港股，US: 美股
        self.backtest_setting = None  # 回测设置






# -*- coding:utf-8 -*-
from collections import OrderedDict
from py_common_util.common.common_utils import CommonUtils


class EnhancedOrderedDict(OrderedDict):
    """
    增强功能的OrderedDict，支持subset功能
    """
    def subset(self, items=[]):
        """返回subset的deep copy对象"""
        if items is None or len(items) < 1:
            return self.copy()
        all = super(EnhancedOrderedDict, self).items()
        # return type(self)((key, value) for (key, value) in all if key in items)
        copied_order_dict = EnhancedOrderedDict()
        for key, value in all:
            if key in items:
                copied_order_dict[key] = CommonUtils.deepcopy(value)
        return copied_order_dict

    def to_list(self):
        """convert OrderedDict's items to list"""
        return list(self.items())

    def next_key(self, key):
        all_keys = list(self.keys())
        if key in all_keys:
            key_index = all_keys.index(key)
        if key_index+1 <= len(all_keys) - 1:
            return all_keys[key_index + 1]
        else:
            return None

    def prev_key(self, key):
        all_keys = list(self.keys())
        if key in all_keys:
            key_index = all_keys.index(key)
        if key_index-1 >= 0:
            return all_keys[key_index - 1]
        else:
            return None


if __name__ == '__main__':
    mod = EnhancedOrderedDict(banana=3, apple=4, pear=1, orange=2)
    print("apple next_key=" + str(mod.next_key("orange")))
    print("apple prev_key=" + str(mod.prev_key("orange")))
    print(list(mod.keys()).index("pear"))
    print(mod.to_list()[-1][1])
    print(mod.get("apple"))
    print(mod.subset())
    print("####test copy")
    a = ["1", "2"]
    original_mod = EnhancedOrderedDict(banana=a)
    copied_mod = original_mod.subset(["apple", "banana"])
    a.append("new value")
    print("original_mod", original_mod)
    print("copied_mod", copied_mod)



# coding: utf-8 or # -*- coding: utf-8 -*-
import pandas as pd
from py_common_util.tushare import MySQLUtils


class PandasUtils(object):

    def __init__(self):
        self._mysql_util = MySQLUtils()
        pass

    def read_sql(self, select_sql):
        return pd.read_sql(select_sql, self._mysql_util.engine)


if __name__ == '__main__':
    pandas = PandasUtils()
    df = pandas.read_sql("SELECT code FROM tushare_2017.stock_basics")
    for code in df.code.values:
        print(code)



# -*- coding:utf-8 -*-
"""
用Python从零开始创建区块链 https://learnblockchain.cn/2017/10/27/build_blockchain_by_python/
"""
import hashlib
import json
from time import time
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse
from uuid import uuid4

import requests
from flask import Flask, jsonify, request


class Blockchain:
    def __init__(self):
        self.current_transactions = []
        self.chain = []
        self.nodes = set()

        # 创建创世块
        self.new_block(previous_hash='1', proof=100)

    def register_node(self, address: str) -> None:
        """
        Add a new node to the list of nodes
        :param address: Address of node. Eg. 'http://192.168.0.5:5000'
        """
        parsed_url = urlparse(address)
        self.nodes.add(parsed_url.netloc)

    def valid_chain(self, chain: List[Dict[str, Any]]) -> bool:
        """
        Determine if a given blockchain is valid

        :param chain: A blockchain
        :return: True if valid, False if not
        """

        last_block = chain[0]
        current_index = 1

        while current_index < len(chain):
            block = chain[current_index]
            print(f'{last_block}')
            print(f'{block}')
            print("\n-----------\n")
            # Check that the hash of the block is correct
            if block['previous_hash'] != self.hash(last_block):
                return False

            # Check that the Proof of Work is correct
            if not self.valid_proof(last_block['proof'], block['proof']):
                return False

            last_block = block
            current_index += 1

        return True

    def resolve_conflicts(self) -> bool:
        """
        共识算法解决冲突
        使用网络中最长的链.

        :return:  如果链被取代返回 True, 否则为False
        """

        neighbours = self.nodes
        new_chain = None

        # We're only looking for chains longer than ours
        max_length = len(self.chain)

        # Grab and verify the chains from all the nodes in our network
        for node in neighbours:
            response = requests.get(f'http://{node}/chain')

            if response.status_code == 200:
                length = response.json()['length']
                chain = response.json()['chain']

                # Check if the length is longer and the chain is valid
                if length > max_length and self.valid_chain(chain):
                    max_length = length
                    new_chain = chain

        # Replace our chain if we discovered a new, valid chain longer than ours
        if new_chain:
            self.chain = new_chain
            return True

        return False

    def new_block(self, proof: int, previous_hash: Optional[str]) -> Dict[str, Any]:
        """
        生成新块
        :param proof: The proof given by the Proof of Work algorithm
        :param previous_hash: Hash of previous Block
        :return: New Block
        """

        block = {
            'index': len(self.chain) + 1,
            'timestamp': time(),
            'transactions': self.current_transactions,
            'proof': proof,
            'previous_hash': previous_hash or self.hash(self.chain[-1]),
        }

        # Reset the current list of transactions
        self.current_transactions = []

        self.chain.append(block)
        return block

    def new_transaction(self, sender: str, recipient: str, amount: int) -> int:
        """
        生成新交易信息，信息将加入到下一个待挖的区块中

        :param sender: Address of the Sender
        :param recipient: Address of the Recipient
        :param amount: Amount
        :return: The index of the Block that will hold this transaction
        """
        self.current_transactions.append({
            'sender': sender,
            'recipient': recipient,
            'amount': amount,
        })

        return self.last_block['index'] + 1

    @property
    def last_block(self) -> Dict[str, Any]:
        return self.chain[-1]

    @staticmethod
    def hash(block: Dict[str, Any]) -> str:
        """
        生成块的 SHA-256 hash值
        :param block: Block
        """
        # We must make sure that the Dictionary is Ordered, or we'll have inconsistent hashes
        block_string = json.dumps(block, sort_keys=True).encode()
        return hashlib.sha256(block_string).hexdigest()

    def proof_of_work(self, last_proof: int) -> int:
        """
        简单的工作量证明:
         - 查找一个 p' 使得 hash(pp') 以4个0开头
         - p 是上一个块的证明,  p' 是当前的证明
        """
        proof = 0
        while self.valid_proof(last_proof, proof) is False:
            proof += 1

        return proof

    @staticmethod
    def valid_proof(last_proof: int, proof: int) -> bool:
        """
        验证证明: 是否hash(last_proof, proof)以4个0开头

        :param last_proof: Previous Proof
        :param proof: Current Proof
        :return: True if correct, False if not.
        """

        guess = f'{last_proof}{proof}'.encode()
        guess_hash = hashlib.sha256(guess).hexdigest()
        return guess_hash[:4] == "0000"


# Instantiate the Node
app = Flask(__name__)

# Generate a globally unique address for this node
node_identifier = str(uuid4()).replace('-', '')

# Instantiate the Blockchain
blockchain = Blockchain()


@app.route('/mine', methods=['GET'])
def mine():
    # We run the proof of work algorithm to get the next proof...
    last_block = blockchain.last_block
    last_proof = last_block['proof']
    proof = blockchain.proof_of_work(last_proof)

    # 给工作量证明的节点提供奖励.
    # 发送者为 "0" 表明是新挖出的币
    blockchain.new_transaction(
        sender="0",
        recipient=node_identifier,
        amount=1,
    )

    # Forge the new Block by adding it to the chain
    block = blockchain.new_block(proof, None)

    response = {
        'message': "New Block Forged",
        'index': block['index'],
        'transactions': block['transactions'],
        'proof': block['proof'],
        'previous_hash': block['previous_hash'],
    }
    return jsonify(response), 200


@app.route('/transactions/new', methods=['POST'])
def new_transaction():
    values = request.get_json()

    # 检查POST数据
    required = ['sender', 'recipient', 'amount']
    if not all(k in values for k in required):
        return 'Missing values', 400

    # Create a new Transaction
    index = blockchain.new_transaction(values['sender'], values['recipient'], values['amount'])

    response = {'message': f'Transaction will be added to Block {index}'}
    return jsonify(response), 201


@app.route('/chain', methods=['GET'])
def full_chain():
    response = {
        'chain': blockchain.chain,
        'length': len(blockchain.chain),
    }
    return jsonify(response), 200


@app.route('/nodes/register', methods=['POST'])
def register_nodes():
    values = request.get_json()
    nodes = values.get('nodes')
    if nodes is None:
        return "Error: Please supply a valid list of nodes", 400

    for node in nodes:
        blockchain.register_node(node)

    response = {
        'message': 'New nodes have been added',
        'total_nodes': list(blockchain.nodes),
    }
    return jsonify(response), 201


@app.route('/nodes/resolve', methods=['GET'])
def consensus():
    replaced = blockchain.resolve_conflicts()

    if replaced:
        response = {
            'message': 'Our chain was replaced',
            'new_chain': blockchain.chain
        }
    else:
        response = {
            'message': 'Our chain is authoritative',
            'chain': blockchain.chain
        }

    return jsonify(response), 200


if __name__ == '__main__':
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('-p', '--port', default=5003, type=int, help='port to listen on')
    args = parser.parse_args()
    port = args.port
    app.run(host='127.0.0.1', port=port)



# coding: utf-8 or # -*- coding: utf-8 -*-
import requests
import time
import traceback
from urllib.parse import urlparse

try:
    # 禁用安全请求警告
    from requests.packages.urllib3.exceptions import InsecureRequestWarning
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
except ImportError:
    print("WARN：禁用安全请求警告操作错误(InsecureRequestWarning import error)")


class RequestsUtils:

    def __init__(self, headers=None, referer="", timeout=(3.05, 5), proxies=[{}]):
        """
         支持动态切换代理(shadowsocks+polipo)
         e.g. proxies=[{},{"http":"http://127.0.0.1:1087","https":"http://127.0.0.1:1087"}，{"http":"http://127.0.0.1:1088","https":"http://127.0.0.1:1088"}]
        """
        if headers is None:
            headers = {
                "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
                "User-Agent": "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.154 Safari/537.36",
                "X-Requested-With": "xmlHttpRequest",
                "Referer": referer,
                "Accept": "*/*",
                'Cache-Control': 'no-cache',
                'If-Modified-Since': '0'
            }
        from py_common_util.common.common_utils import CommonUtils
        if not isinstance(proxies, list):
            proxies = [proxies]
        self._proxies_num = len(proxies)
        self.proxies_iter = CommonUtils.cycle(proxies)
        self.sess = requests.Session()
        self.sess.proxies = next(self.proxies_iter)
        self.sess.verify = False
        self.headers = headers
        self.timeout = timeout
        self.cookies_jar = requests.cookies.RequestsCookieJar()

    def get(self, url, params=None, retry_count=0, retry_count_max=0, refresh_rate_in_second=0, hooks_exception_callback=None):
        """
        get请求, 异常请求则返回None
        :param url:
        :param params:
        :param retry_count:
        :param retry_count_max: 一般为proxies列表的长度减去1
        :param refresh_rate_in_second:
        :param hooks_exception_callback:
        :return:
        """
        try:
            if self.headers.get("Host") is None or len(self.headers.get("Host")) < 1:
                self.headers["Host"] = self.url_parse(url).hostname
            return self.sess.get(url, params=params, headers=self.headers, cookies=self.cookies_jar, timeout=self.timeout)
        except Exception as expt:
            if retry_count < retry_count_max:
                if 'Max retries exceeded with url' in traceback.format_exc():
                    print('出现Max retries exceeded with url异常，也许需要切换代理变量再重试(HTTP_PROXY,HTTPS_PROXY)！')
                    self.update_proxies()
                retry_count += 1
                print("GET URL[" + url + "]retry times:" + str(retry_count) + " of " + str(retry_count_max))
                if refresh_rate_in_second > 0:
                    time.sleep(refresh_rate_in_second)
                return self.get(url, params, retry_count, retry_count_max, refresh_rate_in_second, hooks_exception_callback)
            elif hooks_exception_callback:
                hooks_exception_callback(expt)
            else:
                # will return None
                traceback.print_exc()

    def post(self, url, data=None, retry_count=0, retry_count_max=0, refresh_rate_in_second=0, hooks_exception_callback=None):
        """
        post请求, 异常请求则返回None
        :param url:
        :param data:
        :param retry_count:
        :param retry_count_max: 一般为proxies列表的长度减去1
        :param refresh_rate_in_second:
        :param hooks_exception_callback:
        :return:
        """
        try:
            if self.headers.get("Host") is None or len(self.headers.get("Host")) < 1:
                self.headers["Host"] = self.url_parse(url).hostname
            return self.sess.post(url, data=data, headers=self.headers, cookies=self.cookies_jar, timeout=self.timeout)
        except Exception as expt:
            if retry_count < retry_count_max:
                if 'Max retries exceeded with url' in traceback.format_exc():
                    print('出现Max retries exceeded with url异常，也许需要切换代理变量再重试(HTTP_PROXY,HTTPS_PROXY)！')
                    self.update_proxies()
                retry_count += 1
                print("POST URL[" + url + "]retry times:" + str(retry_count) + " of " + str(retry_count_max))
                if refresh_rate_in_second > 0:
                    time.sleep(refresh_rate_in_second)
                return self.post(url, data, retry_count, retry_count_max, refresh_rate_in_second, hooks_exception_callback)
            elif hooks_exception_callback:
                hooks_exception_callback(expt)
            else:
                # will return None
                traceback.print_exc()

    def close(self):
        self.sess.close()

    def update_proxies(self):
        self.sess.proxies.update(next(self.proxies_iter))

    def get_proxies_num(self):
        return self._proxies_num

    def set_cookie_by_name(self, name, value, domain="", path="/"):
        self.cookies_jar.set(name, value, domain=domain, path=path)

    def remove_cookie_by_name(self, name, domain=None, path=None):
        """Unsets a cookie by name, by default over all domains and paths.

        Wraps CookieJar.clear(), is O(n).
        """
        clearables = []
        for cookie in self.cookies_jar:
            if cookie.name != name:
                continue
            if domain is not None and domain != cookie.domain:
                continue
            if path is not None and path != cookie.path:
                continue
            clearables.append((cookie.domain, cookie.path, cookie.name))

        for domain, path, name in clearables:
            self.cookies_jar.clear(domain, path, name)

    @staticmethod
    def url_parse(url):
        return urlparse(url)

    @staticmethod
    def domain_parse(url, globle_domain=False):
        domain = RequestsUtils.url_parse(url).hostname
        if not globle_domain:
            return domain
        else:
            split_at = domain.split(".")
            i = (0, 1)[globle_domain and len(split_at)>2]
            return ".".join(split_at[i:]).lower()

    @staticmethod
    def decode_content(req_content, encoding="utf-8"):
        return req_content.decode(encoding=encoding)


if __name__ == '__main__':
    def hooks_callback(expt):
        print("这是一个callback函数", expt)
        # raise expt
    url = "http://www.stackoverflow.com?1=2"
    base_url = RequestsUtils.domain_parse(url, True)
    print(base_url)
    parse = RequestsUtils.url_parse('https://api.github.com/events')
    print(parse.hostname)
    req = RequestsUtils(proxies=[{"http":"http://127.0.0.1:10871","https":"http://127.0.0.1:10871"},{"http":"http://127.0.0.1:10872","https":"http://127.0.0.1:10872"},{"http":"http://127.0.0.1:1087","https":"http://127.0.0.1:1087"}])
    # req = RequestsUtils(proxies={"http":"http://127.0.0.1:1087","https":"http://127.0.0.1:1087"})
    # req = RequestsUtils(proxies={})
    # r = req.get(url="https://api.github.com/events", retry_count_max=4)
    # print(r.json())
    # r = req.get(url="https://api.github.com/events", retry_count_max=4)
    # print(r.json())
    # r = req.get(url="https://api.github.com/events", retry_count_max=4)
    # print(r.json())
    try:
        r = req.get(url='https://www.facebook.com', retry_count_max=2, refresh_rate_in_second=3, hooks_exception_callback=hooks_callback)
        print("response=", r)
    except:
        print("主动抛出的异常:", traceback.format_exc())
    # r = req.get(url='https://kyfw.12306.cn/otn/leftTicket/queryZ',
    #           retry_count_max=1)
    # print(r.status_code)



